import asyncio
from types import TracebackType
from typing import Any, Literal, Optional, cast

import aiosqlite

from antares_bot.bot_logging import get_logger
from antares_bot.sqlite.creater import TableDeclarer


SqlRowDict = dict[str, Any]
_LOGGER = get_logger(__name__)


INSERT_COMMAND_FORMAT = """INSERT INTO {table} ({columns})
VALUES
{many_values}
"""
INSERT_COMMAND_PART2_FORMAT = """
ON CONFLICT({pks}) DO UPDATE SET {upsert_args}
"""
SELECT_COMMAND_FORMAT = """SELECT {columns} FROM {table}"""
UPDATE_COMMAND_FORMAT = """UPDATE {table} SET {set}"""
DELETE_COMMAND_FORMAT = "DELETE FROM {table}"
WHERE_PART_FORMAT = """ WHERE {where}"""


class DataBasesManager:
    INST: "DataBasesManager" = None  # type: ignore

    @classmethod
    def get_inst(cls):
        if cls.INST is None:
            cls.INST = cls()
        return cls.INST

    def __init__(self) -> None:
        self._registered_databases: dict[str, Database] = {}

    async def shutdown(self):
        databases = self._registered_databases
        self._registered_databases = {}
        task = asyncio.gather(*(db.close() for db in databases.values()))
        await task
        _LOGGER.info("Closed %d databases", len(databases))

    def register_database(self, name: str, db: "Database"):
        self._registered_databases[name] = db

    def remove_database(self, name: str):
        self._registered_databases.pop(name, None)


class TableProxy(TableDeclarer):
    def __init__(self, db: "Database", table_name: str) -> None:
        super().__init__()
        self.db = db
        self.table_name = table_name
        self.primary_keys: list[str] = []

    def declare_col(self, column_name: str, column_type: str, is_primary: bool = False, is_not_null: bool = False, is_unique: bool = False, default: Any = None):
        super().declare_col(column_name, column_type, is_primary, is_not_null, is_unique, default)
        if is_primary:
            self.primary_keys.append(column_name)

    def _get_parsed_where(self, pk_data: tuple | Any):
        if isinstance(pk_data, tuple):
            if len(pk_data) != len(self.primary_keys):
                raise ValueError("Primary key length not match")
            where = {k: v for k, v in zip(self.primary_keys, pk_data)}
        else:
            where = {self.primary_keys[0]: pk_data}
        return where

    async def _aget_internal(self, select_interface, where):
        rows = await select_interface(self.table_name, where=where)
        rows_list = [dict(row) for row in rows]
        if len(rows_list) > 1:
            raise RuntimeError("More than one row found")
        found = bool(rows_list)
        return found, rows_list[0] if found else None

    async def aget(self, pk_data: tuple | Any):
        where = self._get_parsed_where(pk_data)
        return (await self._aget_internal(self.db.select, where))[1]

    async def agetitem(self, pk_data: tuple | Any):
        where = self._get_parsed_where(pk_data)
        found, value = await self._aget_internal(self.db.select, where)
        if not found:
            raise AttributeError(f"Table {self.table_name} has no item with primary key {pk_data}")
        return value

    def __getitem__(self, pk_data: tuple | Any):
        return self.agetitem(pk_data)

    async def aget_nolock(self, pk_data: tuple | Any):
        where = self._get_parsed_where(pk_data)
        _, value = await self._aget_internal(self.db.select_nolock, where)
        return value

    async def agetitem_nolock(self, pk_data: tuple | Any):
        where = self._get_parsed_where(pk_data)
        found, value = await self._aget_internal(self.db.select_nolock, where)
        if not found:
            raise AttributeError(f"Table {self.table_name} has no item with primary key {pk_data}")
        return value

    def _get_parsed_data_dicts(self, pk_data: tuple | Any, value: SqlRowDict):
        insert_value = value.copy()
        if isinstance(pk_data, tuple):
            for k, v in zip(self.primary_keys, pk_data):
                insert_value[k] = v
        else:
            insert_value[self.primary_keys[0]] = pk_data
        return insert_value

    async def _aset_internal(self, insert_interface, insert_value: SqlRowDict):
        return await insert_interface(self.table_name, insert_value)

    async def aset(self, pk_data: tuple | Any, value: SqlRowDict):
        insert_value = self._get_parsed_data_dicts(pk_data, value)
        return await self._aset_internal(self.db.insert, insert_value)

    async def aset_nolock(self, pk_data: tuple | Any, value: SqlRowDict):
        insert_value = self._get_parsed_data_dicts(pk_data, value)
        return await self._aset_internal(self.db.insert_nolock, insert_value)


class Database(object):
    def __init__(self, dbpath: str) -> None:
        self.db_path = dbpath
        self.conn: aiosqlite.Connection | None = None
        self.lock = asyncio.Lock()
        self.table_info: dict[str, TableProxy] | None = None  # table name -> [(column name, type), ...]
        self.dirty_mark = False
        self._cursor = None
        self._last_command_and_args: tuple[str, Any] | None = None

    async def connect(self) -> None:
        """
        Must be called once at init.
        """
        await self.close()
        self.conn = await aiosqlite.connect(self.db_path)
        self.conn.row_factory = aiosqlite.Row
        DataBasesManager.get_inst().register_database(self.db_path, self)
        await self.update_table_info()

    async def close(self) -> None:
        """
        Call this to close the database.
        It will be called automatically when shutdown, so you do not need to call it manually.
        """
        DataBasesManager.get_inst().remove_database(self.db_path)
        if self.conn is not None:
            try:
                await self.conn.close()
            except Exception:
                ...
            self.conn = None

    def get_cur_connection(self) -> aiosqlite.Connection:
        return cast(aiosqlite.Connection, self.conn)

    @property
    def cursor(self) -> aiosqlite.Cursor:
        return cast(aiosqlite.Cursor, self._cursor)

    async def update_table_info(self) -> None:
        # self.cursor maybe None here
        c = await self.get_cur_connection().cursor()
        tables_info = await (await c.execute("select name from sqlite_master where type='table';")).fetchall()
        tables_key: list[str] = [t[0] for t in tables_info]
        self.table_info = dict()
        for table_name in tables_key:
            table_info = await (await c.execute(f"PRAGMA table_info({table_name});")).fetchall()
            tb_declare = TableProxy(self, table_name)
            self.table_info[table_name] = tb_declare
            for row in table_info:
                tb_declare.declare_col(
                    row['name'],
                    row['type'],
                    is_primary=row['pk'] > 0,
                    is_not_null=row['notnull'] > 0,
                    default=row['dflt_value'],
                )

    def get_primary_key_names(self, table: str) -> list[str]:
        assert self.table_info is not None
        return self.table_info[table].primary_keys

    @staticmethod
    def _key_eq_value_format(k, v, parse_arg: list):
        parse_arg.append(v)
        return f"{k}=?"

    async def select_nolock(
        self, table: str, where: SqlRowDict | None = None, need: list[str] | None = None
    ):
        command = SELECT_COMMAND_FORMAT.format(
            table=table,
            columns=",".join(need) if need else "*"
        )
        parse_args: list = []
        if where:
            command += WHERE_PART_FORMAT.format(where=" AND ".join(
                self._key_eq_value_format(k, v, parse_args) for k, v in where.items()
            ))
        command += ";"

        _LOGGER.debug("execute command %s with args: %s", command, parse_args)
        self._last_command_and_args = (command, parse_args)

        await self.cursor.execute(command, parse_args)
        return await self.cursor.fetchall()

    async def insert_nolock(self, table: str, data_dicts: list[SqlRowDict] | SqlRowDict):
        if not isinstance(data_dicts, list):
            data_dicts = [data_dicts]
        if len(data_dicts) == 0:
            return

        # do column name check
        columns = list(data_dicts[0].keys())
        columns_set = set(columns)
        for i in range(1, len(data_dicts)):
            if set(data_dicts[i].keys()) != columns_set:
                raise ValueError("Column name not match")

        pks = self.get_primary_key_names(table)

        one_value = "(" + ",".join(["?" for _ in columns]) + ")"
        many_values = ",\n".join([one_value for _ in data_dicts])

        insert_command = INSERT_COMMAND_FORMAT.format(
            table=table,
            columns=",".join(columns),
            many_values=many_values,
        )
        if 0 < len(pks) < len(columns):
            upsert_args = ','.join([f"{col}=excluded.{col}" for col in columns if col not in pks])
            insert_command += INSERT_COMMAND_PART2_FORMAT.format(
                pks=",".join(pks),
                upsert_args=upsert_args,
            )
        insert_command += ";"

        parse_args:list[Any] = []
        for data_dict in data_dicts:
            parse_args.extend(data_dict[col] for col in columns)

        _LOGGER.debug("execute command %s with args: %s", insert_command, parse_args)
        self._last_command_and_args = (insert_command, parse_args)

        await self.cursor.execute(insert_command, parse_args)
        self.dirty_mark = True

    async def update_nolock(self, table: str, datadict: SqlRowDict, where: SqlRowDict | Literal["*"] | None = None):
        if where:
            where_data: SqlRowDict | None = None if where == "*" else where
        else:
            pks = self.get_primary_key_names(table)
            where_data = dict()
            for k in pks:
                where_data[k] = datadict.pop(k)
        if len(datadict) == 0:
            _LOGGER.debug("nothing to set, no need to update database")
            return

        parse_args: list = []
        command = UPDATE_COMMAND_FORMAT.format(
            table=table,
            set=",".join(self._key_eq_value_format(k, v, parse_args) for k, v in datadict.items()),
        )

        if where_data is not None:
            command += WHERE_PART_FORMAT.format(where=" AND ".join(
                self._key_eq_value_format(k, v, parse_args) for k, v in where_data.items()
            ))

        command += ";"

        _LOGGER.debug("parse command %s with args: %s", command, parse_args)
        self._last_command_and_args = (command, parse_args)

        await self.cursor.execute(command, parse_args)
        self.dirty_mark = True

    async def delete_nolock(self, table: str, where: SqlRowDict | Literal["*"]):
        command = DELETE_COMMAND_FORMAT.format(table=table)
        parse_args: list = []

        if where != "*":
            command += WHERE_PART_FORMAT.format(where=" AND ".join([
                self._key_eq_value_format(k, v, parse_args) for k, v in where.items()
            ]))

        command += ";"

        _LOGGER.debug("parse command %s with args: %s", command, parse_args)
        self._last_command_and_args = (command, parse_args)

        await self.cursor.execute(command, parse_args)
        self.dirty_mark = True

    async def select(
        self,
        table: str,
        where: SqlRowDict | None = None,
        need: list[str] | None = None
    ):
        async with self:
            return await self.select_nolock(table, where, need)

    async def insert(self, table: str, data_dicts: list[SqlRowDict] | SqlRowDict):
        async with self:
            await self.insert_nolock(table, data_dicts)

    async def update(self, table: str, datadict: SqlRowDict, where: SqlRowDict | Literal['*'] | None = None):
        async with self:
            await self.update_nolock(table, datadict, where)

    async def delete(self, table, where: SqlRowDict | Literal['*']):
        async with self:
            await self.delete_nolock(table, where)

    async def execute(self, cmd: list[str], need_commit: bool = True):
        """execute a list of commands."""
        async with self:
            for c in cmd:
                await self.cursor.execute(c)
            # await self.get_cur_connection().commit()
            self.dirty_mark = need_commit

    async def __aenter__(self):
        if self.conn is None:
            raise RuntimeError("Database not connected")
        await self.lock.acquire()
        self._cursor = await self.get_cur_connection().cursor()
        self._last_command_and_args = None
        return True

    async def __aexit__(self, exception_type, exception_value, exception_traceback: Optional["TracebackType"]):
        if exception_type is not None:
            if self._last_command_and_args is not None:
                last_command, last_args = self._last_command_and_args
                _LOGGER.error("Error occurred when executing command %s with args: %s", last_command, last_args)
        if self.dirty_mark:
            try:
                await self.get_cur_connection().commit()
            except Exception as e:
                from antares_bot.utils import exception_manual_handle
                await exception_manual_handle(_LOGGER, e)
            self.dirty_mark = False
        self._cursor = None
        self.lock.release()
        self._last_command_and_args = None
        return False

    def __getitem__(self, table: str) -> TableProxy:
        assert self.table_info
        return self.table_info[table]



from typing import TYPE_CHECKING, Any, Dict, Optional, cast

from telegram import Update
from telegram.constants import ChatType
from telegram.ext import Application, CallbackContext, ExtBot

from antares_bot.utils import ObjectDict, get_msg_id, get_reply_to_msg_id


if TYPE_CHECKING:
    from typing import Self


class ChatData:
    def __init__(self) -> None:
        self.rich_data: Dict[int, Any] = ObjectDict()
        """key: message id"""


class UserData:
    pass


class RichCallbackContext(CallbackContext[ExtBot, UserData, ChatData, dict]):
    def __init__(
            self,
            application: Application,
            chat_id: int | None = None,
            user_id: int | None = None
    ):
        super().__init__(application, chat_id, user_id)
        self._last_message_id: Optional[int] = None
        self._reply_to_message_id: Optional[int] = None
        self._update: Optional[Update] = None
        self._type = ""

    @classmethod
    def from_update(cls, update: object, application: "Application") -> "Self":
        """Override from_update to set _message_id."""
        context = super().from_update(update, application)
        if isinstance(update, Update):
            # pylint: disable=protected-access
            context._last_message_id = get_msg_id(update)
            context._reply_to_message_id = get_reply_to_msg_id(update)
            context._update = update
            if update.effective_chat is not None:
                context._type = update.effective_chat.type
            # pylint: enable=protected-access
        return context

    def is_private_chat(self):
        return self._type == ChatType.PRIVATE

    def is_group_chat(self):
        return self._type in [ChatType.GROUP, ChatType.SUPERGROUP]

    def is_channel_message(self):
        return self._type == ChatType.CHANNEL

    def chat_type_str(self):
        return self._type

    def is_callback_query(self):
        return self._update.callback_query is not None

    def get_key(self):
        return (self.chat_id, self.user_id)

    @property
    def user_id(self) -> int:
        return cast(int, self._user_id)

    @property
    def chat_id(self) -> int:
        return cast(int, self._chat_id)

    @property
    def message_id(self) -> int:
        return cast(int, self._last_message_id)

    @property
    def reply_to_message_id(self) -> int:
        return cast(int, self._reply_to_message_id)

    @property
    def type(self) -> str:
        return self._type



from typing import TYPE_CHECKING, Optional

from antares_bot.bot_default_cfg import BasicConfig
from antares_bot.context_manager import InvalidContext, get_context
from antares_bot.utils import read_user_cfg


if TYPE_CHECKING:
    from antares_bot.context import RichCallbackContext


def get_default_locale() -> str | None:
    return read_user_cfg(BasicConfig, "LOCALE")


def lang_context():
    try:
        ct = get_context()
        if ct is None or isinstance(ct, InvalidContext):
            return get_default_locale()
    except Exception:
        return get_default_locale()
    return LangContextManager.get_inst().get_user_lang(ct)


def set_lang(_id: int, locale: str):
    # id < 0 -> group, id > 0 -> private
    LangContextManager.get_inst().set_user_lang(_id, locale)


class LangContextManager:
    INST: Optional["LangContextManager"] = None

    def __init__(self) -> None:
        self.user_lang_context: dict[int, str] = dict()

    def get_user_lang(self, ct: "RichCallbackContext") -> str | None:
        return self.user_lang_context.get(ct.user_id, get_default_locale())  # type: ignore

    def set_user_lang(self, _id: int, locale: str):
        self.user_lang_context[_id] = locale

    @classmethod
    def get_inst(cls):
        if cls.INST is None:
            cls.INST = LangContextManager()
        return cls.INST



import os
import sys
from typing import TYPE_CHECKING, Awaitable, Callable


if TYPE_CHECKING:
    from antares_bot.bot_inst import TelegramBot


ANTARES_BOT_LOG_LEVEL_DEFAULT = "WARN"


def bootstrap(
    from_cmd_line=False,
    custom_post_init: Callable[[], Awaitable] | None = None,
    **kwargs
) -> "TelegramBot":
    """
    Bootstrap the bot.
    If from_cmd_line is True, it will parse the command line arguments.
    Otherwise it will use the kwargs.
    Supported kwargs: force_pika, log_level
    """
    # phase 1
    if from_cmd_line:
        # examine if working directory is in sys.path
        cwd = os.path.normpath(os.path.abspath(os.getcwd()))
        # if not, add to first. only do this if called from cmd line
        if cwd not in list(map(lambda x: os.path.normpath(os.path.abspath(x)), sys.path)):
            sys.path = [cwd] + sys.path

    # phase 2
    # define arg: force_pika and log_level
    if from_cmd_line:
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument("--update-pika-interface", action="store_true", default=False,
                            help="Force download latest Pika interface from GitHub", required=False, dest="force_pika")
        parser.add_argument("--log-level", type=str, default=ANTARES_BOT_LOG_LEVEL_DEFAULT, help="Logging level", required=False)
        args = parser.parse_args()
        force_pika: bool = args.force_pika
        log_level = args.log_level
    else:
        force_pika = kwargs.get("force_pika", False)
        log_level = kwargs.get("log_level", ANTARES_BOT_LOG_LEVEL_DEFAULT)

    # phase 3
    # pika init phase. This is optional.
    def pika_init(_force_pika: bool) -> None:
        # cannot import init related things here
        from antares_bot.bot_default_cfg import AntaresBotConfig
        from antares_bot.init_hooks import init_pika, read_user_cfg
        # When init_hooks is imported, the cfg is automatically hooked.
        # Now we can call read_user_cfg safely.
        skip_pika_setup = False
        if not _force_pika:
            skip_pika_setup = read_user_cfg(AntaresBotConfig, "SKIP_PIKA_SETUP") or False
        if _force_pika or not skip_pika_setup:
            init_pika(force_update=_force_pika)
        # logging can start after pika inited.
        # As long as the logging module is imported, it is inited.
    pika_init(_force_pika=force_pika)

    import logging
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        level=log_level,
    )

    # phase 4
    # initiate bot
    from antares_bot.bot_inst import get_bot_instance
    bot_app = get_bot_instance()

    # phase 5
    # setup post_init. This is optional.
    if custom_post_init:
        bot_app.custom_post_init(custom_post_init())

    # DONE
    return bot_app


def main():
    bot_app = bootstrap(from_cmd_line=True)
    # start!
    bot_app.run()
    # exit
    return 0



from typing import cast

from antares_bot.bot_logging import get_logger
from antares_bot.multi_lang.context import get_default_locale, lang_context


_LOGGER = get_logger(__name__)

_CACHE_LOGGED: set[tuple[int, str | None]] = set()


class LocaleError(ValueError):
    pass


def t(d: dict[str, str], locale: str | None = None) -> str:
    """
    Translate message by language context.
    May raise if the locale augument is not `None` and invalid.
    If locale is None, we use the default language.
    If the default language does not appear in `d`, use the first
    dict value of `d`. Will not raise any error
    as long as the language dict `d` is not empty.
    """
    l_ct = lang_context() if locale is None else locale
    ans: str | None = cast(str | None, d.get(l_ct, None))
    if ans is None:
        if locale is not None:
            # specified a locale but not found. Should raise an error
            raise LocaleError(f"Locale {locale} not found in {d}")
        ans = cast(str | None, d.get(get_default_locale(), None))  # type: ignore
        if ans is None:
            # give the first value
            if (id(d), l_ct) not in _CACHE_LOGGED:
                _CACHE_LOGGED.add((id(d), l_ct))
                _LOGGER.warning("No text found in %s for locale %s, using the first value", d, l_ct)
            ans = next(iter(d.values()))
    return ans



from logging import DEBUG as LOGLEVEL_DEBUG
from typing import TYPE_CHECKING, Any, Optional

from telegram import InlineKeyboardButton
from telegram.error import TelegramError

import antares_bot.context_manager as context_manager
from antares_bot import utils
from antares_bot.bot_default_cfg import BasicConfig
from antares_bot.bot_logging import get_logger, get_root_logger
from antares_bot.bot_method_wrapper import TelegramBotBaseWrapper
from antares_bot.error import IgnoreChannelUpdateException, InvalidChatTypeException, UserPermissionException
from antares_bot.permission_check import CheckLevel, ConditionLimit, PermissionState, permission_check


if TYPE_CHECKING:
    from telegram import Update

    from antares_bot.context import RichCallbackContext

DEBUG_LOGGER = get_logger("debug")


class TelegramBotBase(TelegramBotBaseWrapper):
    @classmethod
    def check(cls, level: CheckLevel, limit: ConditionLimit = ConditionLimit.ALL) -> None:
        check_result = permission_check(cls.get_context(), level, limit)
        cls._check_raise(check_result)

    @classmethod
    def _check_raise(cls, check_result: PermissionState):
        if check_result == PermissionState.PASSED:
            return
        if check_result == PermissionState.INVALID_USER:
            raise UserPermissionException
        elif check_result == PermissionState.INVALID_CHAT_TYPE:
            raise InvalidChatTypeException
        elif check_result == PermissionState.IGNORE_CHANNEL:
            raise IgnoreChannelUpdateException

    @classmethod
    def get_context(cls) -> "RichCallbackContext":
        context = cls.peek_context()
        if context is None:
            raise RuntimeError("No context found")
        return context

    @classmethod
    def peek_context(cls) -> Optional["RichCallbackContext"]:
        return context_manager.get_context()

    ##############################

    @classmethod
    async def del_msg(cls, chat_id: int, msgid: int, maxTries: int = 5) -> bool:
        from antares_bot.bot_inst import get_bot_instance
        if maxTries <= 0:
            raise ValueError("Invalid retry times")

        for i in range(maxTries):
            try:
                await get_bot_instance().bot.delete_message(chat_id=chat_id, message_id=msgid)
            except TelegramError:
                if i == maxTries - 1:
                    return False
                continue
            break

        return True

    @classmethod
    def debug_info(cls, msg, *args):
        DEBUG_LOGGER.debug(msg, *args)

    ##############################

    @classmethod
    def get_master_id(cls) -> int:
        return utils.read_user_cfg(BasicConfig, "MASTER_ID")

    @classmethod
    def is_master(cls, ct: "RichCallbackContext | None" = None) -> bool:
        # TODO: consider private channel
        master_id = cls.get_master_id()
        if ct is None:
            ct = cls.get_context()
        return ct.chat_id == master_id or ct.user_id == master_id

    @classmethod
    def format_inline_keyboard_button(cls, text: str, callback_key_name: str, callback_arg: Any):
        return InlineKeyboardButton(text, callback_data=f"{callback_key_name}:{callback_arg}")

    @classmethod
    def _is_debug_level(cls):
        root_logger = get_root_logger()
        return root_logger.level == LOGLEVEL_DEBUG

    @classmethod
    async def fetch_url(cls, url: str):
        return await utils.fetch_url(url)

    @classmethod
    def get_message_after_command(cls, update: "Update") -> str:
        assert update.message is not None
        assert update.message.text is not None
        txt = update.message.text.strip()
        pre_command = txt.split()[0]
        return txt[len(pre_command):].strip()

    @property
    def bot_id(self):
        raise NotImplementedError



import os
from typing import Any, Dict

import aiosqlite

from antares_bot.bot_logging import get_logger


_LOGGER = get_logger(__name__)

INT = "INT"
TEXT = "TEXT"
REAL = "REAL"
BLOB = "BLOB"
NUMERIC = "NUMERIC"


class NoTableException(Exception):
    pass


class NoDbPathException(Exception):
    pass


class NoColumnException(Exception):
    pass


class DbCreationException(Exception):
    pass


class ColumnDeclarer(object):
    def __init__(self, column_name: str, column_type: str, is_primary: bool = False, is_not_null: bool = False, is_unique: bool = False, default: Any = None) -> None:
        self.column_name = column_name
        self.column_type = column_type
        self.is_primary = is_primary
        self.is_not_null = is_not_null
        self.is_unique = is_unique
        self.default = default


class TableDeclarer(object):
    def __init__(self) -> None:
        self.table_name = ""
        self.columns: Dict[str, ColumnDeclarer] = dict()
        self.pkey_count = 0

    def set_table_name(self, table_name):
        self.table_name = table_name
        return self

    def declare_col(self, column_name: str, column_type: str, is_primary: bool = False, is_not_null: bool = False, is_unique: bool = False, default: Any = None):
        if is_primary:
            self.pkey_count += 1
        self.columns[column_name] = ColumnDeclarer(column_name, column_type, is_primary, is_not_null, is_unique, default)
        return self

    def get_creation_cmd(self):
        """
        Returns the SQL string for creating the table based on the specified table name and columns.

        Raises:
            NoTableException: If the table name is not declared.
            NoColumnException: If no columns are declared for the table.

        Returns:
            str: The SQL string for creating the table.
        """
        if self.table_name == "":
            raise NoTableException("Table name not declared")
        if len(self.columns) == 0:
            raise NoColumnException("No column declared: {}".format(self.table_name))
        execute_str = "CREATE TABLE {} (".format(self.table_name)
        for column in self.columns.values():
            execute_str += "\n{} {} ".format(column.column_name, column.column_type)
            if column.is_primary and self.pkey_count == 1:
                execute_str += "PRIMARY KEY "
            if column.is_not_null:
                execute_str += "NOT NULL "
            if column.is_unique:
                execute_str += "UNIQUE "
            if column.default is not None:
                if column.column_type == TEXT:
                    execute_str += "DEFAULT '{}' ".format(column.default)
                else:
                    execute_str += "DEFAULT {} ".format(column.default)
            execute_str += ","
        if self.pkey_count > 1:
            execute_str = execute_str + "\nPRIMARY KEY ("
            for column in self.columns.values():
                if column.is_primary:
                    execute_str += column.column_name + ","
            execute_str = execute_str[:-1] + "));"
        else:
            execute_str = execute_str[:-1] + ");"
        return execute_str

    def get_data_creator(self):
        return TableDataCreator(self)


class TableDataCreator(object):
    def __init__(self, table: TableDeclarer):
        self.table = table
        self.data: Dict[str, Any] = dict()

    def set(self, column: str, value: Any):
        if column not in self.table.columns:
            raise ValueError("column {} not declared".format(column))
        self.data[column] = value
        return self

    def set_dict(self, data: Dict[str, Any]):
        for k in data:
            if k not in self.table.columns:
                raise ValueError("column {} not declared".format(k))
        self.data.update(data)
        return self

    def create(self):
        for col in self.table.columns.values():
            if col.column_name not in self.data:
                if col.is_not_null and col.default is None:
                    raise RuntimeError("column {} is not nullable".format(col.column_name))
        return self.data


class DbDeclarer(object):
    def __init__(self) -> None:
        self.db_path = ""
        self.tables: Dict[str, TableDeclarer] = dict()

    def declare(self, db_path: str):
        self.db_path = db_path
        return self

    def declare_table(self, table_name: str):
        t = TableDeclarer()
        self.tables[table_name] = t.set_table_name(table_name)
        return t

    async def create_or_validate(self):
        if os.path.exists(self.db_path):
            await self.validate()
        else:
            await self.create()

    async def create(self):
        if self.db_path == "":
            raise NoDbPathException("Database path not declared")
        dir_name = os.path.dirname(self.db_path)
        os.makedirs(dir_name, exist_ok=True)
        if len(self.tables) == 0:
            raise NoTableException("No table declared: {}".format(self.db_path))
        try:
            conn = await aiosqlite.connect(self.db_path)
            c = await conn.cursor()
            # drop table if exists
            for table_name in self.tables:
                command = "DROP TABLE IF EXISTS {}".format(table_name)
                _LOGGER.warning(command)
                await c.execute(command)
            for table in self.tables.values():
                command = table.get_creation_cmd()
                _LOGGER.warning(command)
                await c.execute(command)
            await conn.commit()
            await conn.close()
        except Exception as e:
            raise DbCreationException from e

    async def validate(self):
        # check if table exists
        conn = await aiosqlite.connect(self.db_path)
        c = await conn.cursor()
        for table in self.tables.values():
            command = "SELECT name FROM sqlite_master WHERE type='table' AND name='{}'".format(table.table_name)
            _LOGGER.debug(command)
            await c.execute(command)
            if not await c.fetchone():
                _LOGGER.warning("Table %s not exists", table.table_name)
                command = table.get_creation_cmd()
                _LOGGER.warning(command)
                await c.execute(command)
        await conn.commit()
        await conn.close()



import asyncio
import os
import tempfile
from io import StringIO
from typing import TYPE_CHECKING, List, Union

import objgraph  # type: ignore

from antares_bot.bot_default_cfg import AntaresBotConfig
from antares_bot.framework import command_callback_wrapper
from antares_bot.module_base import TelegramBotModuleBase
from antares_bot.permission_check import CheckLevel
from antares_bot.utils import read_user_cfg


if TYPE_CHECKING:
    from telegram import Update
    from telegram.ext import BaseHandler

    from antares_bot.context import RichCallbackContext
    from antares_bot.framework import CallbackBase


class ObjGraph(TelegramBotModuleBase):
    def do_init(self) -> None:
        if not read_user_cfg(AntaresBotConfig, "OBJGRAPH_TRACE_AT_START"):
            return
        s = StringIO()
        objgraph.show_most_common_types(limit=50, file=s)
        objgraph.show_growth()  # ignore return value
        ss = s.getvalue()

        async def _f(context: "RichCallbackContext"):
            await self.send_to(self.get_master_id(), ss)
        self.job_queue.run_once(_f, 10)

    def mark_handlers(self) -> List[Union["CallbackBase", "BaseHandler"]]:
        return [self.memory_graph, self.backrefs, self.tracemalloc]

    @command_callback_wrapper
    async def memory_graph(self, update: "Update", context: "RichCallbackContext"):
        self.check(level=CheckLevel.MASTER)
        s1 = StringIO()
        s2 = StringIO()
        objgraph.show_most_common_types(limit=50, file=s1)
        objgraph.show_growth(limit=50, file=s2)
        await self.reply("show_most_common_types:\n" + s1.getvalue())
        await self.reply("show_growth:\n" + s2.getvalue())

    @command_callback_wrapper
    async def backrefs(self, update: "Update", context: "RichCallbackContext"):
        """
        `/backrefs <ObjType>`: Draw a graph of the objects that reference the objects of given type.
        """
        self.check(level=CheckLevel.MASTER)

        if not context.args:
            await self.error_info("Usage: /backrefs <ObjType>")
            return
        temp_dir = tempfile.gettempdir()
        objs = objgraph.by_type(context.args[0])
        if len(objs) == 0:
            await self.error_info("No objects found")
            return
        loop = asyncio.get_running_loop()
        loop.create_task(self.reply(f"found {len(objs)} objects in memory"))
        file_name = os.path.join(temp_dir, "backrefs.svg")
        objgraph.show_backrefs(objs, max_depth=5, filename=file_name)
        del objs
        await self.reply_document(file_name)
        os.remove(file_name)

    @command_callback_wrapper
    async def tracemalloc(self, update: "Update", context: "RichCallbackContext"):
        self.check(level=CheckLevel.MASTER)
        assert context.args is not None
        is_stop = len(context.args) > 0 and context.args[0] == "stop"
        import tracemalloc
        if is_stop:
            tracemalloc.stop()
            await self.reply("tracemalloc stopped")
            return
        if tracemalloc.is_tracing():
            snapshot = tracemalloc.take_snapshot()
            top_stats = snapshot.statistics('lineno')
            loop = asyncio.get_running_loop()
            loop.create_task(self.reply(f"There is {len(top_stats)} tracemalloc stats"))
            top_stats = top_stats[:200]
            stats_str = '\n'.join(str(stat) for stat in top_stats)
            await self.reply(stats_str)
            return
        else:
            tracemalloc.start()
            await self.reply("tracemalloc started")
            return



# should not import anything other than python stdlib here
import os
import sys


_BLANK_CFG = \
    """\"\"\"Naming convention: as long as the config item can be retrieved like:
```
bot_cfg.<SectionName>.<ITEM_NAME>
```
it is valid. For example: `bot_cfg.BasicConfig.TOKEN`
The `BasicConfig` and `AntaresBotConfig` are required.
\"\"\"


class BasicConfig:
    \"\"\"
    Basic sample config.
    TOKEN and MASTER_ID must be provided.
    For more details, refers to bot_default_cfg.py.
    \"\"\"
    # LOCALE = 'en'
    # DATA_DIR = "data"
    TOKEN = "abcdef:123456"
    MASTER_ID = 123456789
    # BOT_NAME = "my_bot"  # mostly used in logging


class AntaresBotConfig:
    \"\"\"
    Config to control the bot behavior. They are all optional.
    For more details, refers to bot_default_cfg.py.
    \"\"\"
    PULL_WHEN_STOP = True  # enable pulling when bot stops
    # JOB_QUEUE_CONFIG = {
    #     "job_defaults" : {
    #         "misfire_grace_time": 30  # default is 60
    #     }
    # }
    # SKIP_PIKA_SETUP = True
    # SKIP_LOAD_ALL_INTERNAL_MODULES = True
    # SKIP_LOAD_ALL_MODULES = True
    # SKIP_LOAD_INTERNAL_MODULE_{module_name_upper_case} = True
    # SKIP_LOAD_MODULE_{module_name_upper_case} = True
    # OBJGRAPH_TRACE_AT_START = True
    # SYSTEMD_SERVICE_NAME = "antares_bot.service"
    # IGNORE_IMPORT_MODULE_ERROR = True
"""


def _hook_cfg():
    try:
        import bot_cfg as cfg  # type: ignore
    except ImportError:
        print(
            "./bot_cfg.py not found or cannot be imported, I will try to create one for you.",
            file=sys.stderr
        )
        create_blank_cfg()
        exit(1)
    except SyntaxError:
        print(
            "./bot_cfg.py has syntax error, please fix it before running.",
        )
        exit(1)
    from antares_bot import bot_default_cfg
    BaseConfig = bot_default_cfg.BaseConfig
    # start checking
    for k in dir(bot_default_cfg):
        default_config_class = getattr(bot_default_cfg, k)
        empty_checklist = getattr(default_config_class, 'non_empty', [])
        if isinstance(default_config_class, type) and issubclass(default_config_class, BaseConfig) and default_config_class != BaseConfig:
            if not hasattr(cfg, k):
                if len(empty_checklist) != 0:
                    print(
                        f"Please create {default_config_class.__name__} in bot_cfg.py and fill in the required fields first."
                        " See bot_default_cfg.py for more details."
                        " Or, you can remove the bot_cfg.py and run again to create a new one",
                        file=sys.stderr
                    )
                    exit(1)
                setattr(cfg, k, default_config_class)
                config_class = default_config_class
            else:
                config_class = getattr(cfg, k)
                for attr in default_config_class.iter_all_config_keys():
                    if not hasattr(config_class, attr):
                        if attr in empty_checklist:
                            print(
                                f"Please set {', '.join(empty_checklist)} in "
                                f"{default_config_class.__name__} in bot_cfg.py. See bot_default_cfg.py for more details.",
                                file=sys.stderr
                            )
                            exit(1)
                        setattr(config_class, attr, getattr(default_config_class, attr))


def init_pika(force_update: bool = False):
    # for better check
    gh_username = 'Antares0982'
    gh_repo = 'PikaInterface'
    gh_file = 'pika_interface.py'
    if not force_update and os.path.exists(gh_file):
        print(f"{gh_file} already exists, skipping download")
        return
    if force_update:
        print(f"Force updating {gh_file}")
    else:
        # read stdin to ensure the user is aware of the download
        read_input = input(
            "This operation will download the latest version of"
            f"{gh_file} from GitHub (Please note that using pika for logging also requires installing aio-pika, e.g. `pip install aio-pika`).\n"
            f"You may want to check the source code here: https://github.com/{gh_username}/{gh_repo}/blob/main/{gh_file}.\n"
            "Do you want to continue? (y/N)"
        )
        if not read_input.lower().strip().startswith('y'):
            print("Download skipped. If you want to skip this step forever, set SKIP_PIKA_SETUP in AntaresBotConfig to True.")
            # make sure the user is aware of this message
            import time
            time.sleep(2)
            return
    import subprocess
    command = f'curl -O https://raw.githubusercontent.com/{gh_username}/{gh_repo}/main/{gh_file}'
    print(command)
    code = subprocess.call(command, shell=True)
    if code != 0:
        print(f"failed to download {gh_file}", file=sys.stderr)
        return
    print(f"{gh_file} downloaded successfully.")


def read_user_cfg(cfg_class, section: str):
    import bot_cfg  # type: ignore
    class_name = cfg_class.__name__
    cfg = getattr(bot_cfg, class_name, None)
    return None if cfg is None else getattr(cfg, section, None)


def create_blank_cfg():
    try:
        with open("bot_cfg.py", "x", encoding="utf-8", opener=lambda x, y: os.open(x, y, 0o600)) as f:
            f.write(_BLANK_CFG)
    except FileExistsError:
        print(
            "bot_cfg.py already exists but cannot be imported, skipping creation.\n"
            "Please remove/modify it before running.",
            file=sys.stderr
        )
        return
    print(
        "bot_cfg.py created successfully."
        " Please fill in the required fields."
    )


_hook_cfg()



from typing import TYPE_CHECKING

from antares_bot.module_base import TelegramBotModuleBase
from antares_bot.framework import command_callback_wrapper

if TYPE_CHECKING:
    from telegram import Update

    from antares_bot.context import RichCallbackContext


class Echo(TelegramBotModuleBase):
    def mark_handlers(self):
        return [self.echo]

    @command_callback_wrapper
    async def echo(self, update: "Update", context: "RichCallbackContext") -> bool:
        assert update.message is not None
        assert update.message.text is not None
        text = update.message.text.strip()
        if text.startswith("/echo"):
            text = text[len("/echo"):].strip()
        if text:
            await self.reply(text)
        return True



class BasicLanguage:
    CANCELLED = {
        "zh-CN": "操作取消～",
        "en": "Cancelled~",
    }
    END = {
        "zh-CN": "结束",
        "en": "END",
    }
    INVALID_CHAT_TYPE = {
        "zh-CN": "不能在这里使用哦",
        "en": "Cannot be used in this chat type",
    }
    NO_PERMISSION = {
        "zh-CN": "你没有权限哦",
        "en": "Oops, you don't have permission...",
    }
    SHORTSEP = {
        "zh-CN": "、",
        "en": ", ",
    }
    STARTUP = {
        "zh-CN": "Bot启动！",
        "en": "Bot is live!",
    }
    UNKNOWN_ERROR = {
        "zh-CN": "哎呀，出现了未知的错误呢……",
        "en": "Oops, an unknown error occurred...",
    }
    SHUTDOWN_GOODBYTE = {
        "zh-CN": "主人再见QAQ",
        "en": "Goodbye master QAQ",
    }
    DEBUG_MODE_ON = {
        "zh-CN": "已开启debug模式",
        "en": "Debug mode on",
    }
    DEBUG_MODE_OFF = {
        "zh-CN": "已关闭debug模式",
        "en": "Debug mode off",
    }
    EXEC_FAILED = {
        "zh-CN": "执行失败……",
        "en": "Execution failed...",
    }
    EXEC_SUCCEEDED = {
        "zh-CN": "执行成功，返回值：\n```\n{}\n```",
        "en": "Execution succeeded, return value:\n```\n{}\n```",
    }
    NO_EXEC_COMMAND = {
        "zh-CN": "没有接收到命令诶",
        "en": "No executable command received",
    }
    FORWARD_MESSAGE_FROM_CHANNEL = {
        "zh-CN": "转发消息的来源频道id：`{}`",
        "en": "Forwarded from channel id: {}",
    }
    FORWARD_MESSAGE_FROM_GROUP = {
        "zh-CN": "转发消息的来源群id：`{}`",
        "en": "Forwarded from group id: `{}`",
    }
    FORWARD_MESSAGE_FROM_USER = {
        "zh-CN": "转发消息的来源用户id：`{}`",
        "en": "Forwarded from user id: `{}`",
    }
    FORWARD_FROM_HIDDEN_USER = {
        "zh-CN": "转发消息的来源用户已隐藏",
        "en": "Forwarded from a hidden user",
    }
    GROUP_ID = {
        "zh-CN": "群id：`{}`",
        "en": "Group id: `{}`",
    }
    REPLY_MESSAGE_USER_ID = {
        "zh-CN": "回复消息的用户id：`{}`",
        "en": "Reply message user id: `{}`",
    }
    USER_ID = {
        "zh-CN": "您的id：`{}`",
        "en": "Your id: `{}`",
    }
    NO_SUCH_COMMAND = {
        "zh-CN": "没有找到命令：{}",
        "en": "No such command: {}",
    }

    @classmethod
    def t(cls, d: dict[str, str], locale: str | None = None):
        """
        Translate message by language context.
        May raise if the locale augument is invalid.
        If locale is None, we use the default context, will not raise any error
        as long as the language dict `d` is not empty.
        """
        from antares_bot.multi_lang import t
        return t(d, locale)



import datetime
import os
from typing import TYPE_CHECKING, List, Optional, cast

import httpx
from telegram import InlineKeyboardButton, InlineKeyboardMarkup, Update
from telegram.constants import ChatType

from antares_bot.bot_default_cfg import AntaresBotConfig
from antares_bot.init_hooks import read_user_cfg


if TYPE_CHECKING:
    import logging

SYSTEM_TIME_ZONE = cast(datetime.tzinfo, datetime.datetime.now().astimezone().tzinfo)


class ObjectDict(dict):
    """
    General json object that allows attributes 
    to be bound to and also behaves like a dict.
    """

    def __getattr__(self, attr: str):
        return self.get(attr)

    def __setattr__(self, attr: str, value):
        self[attr] = value


class ChatMessage(object):
    __slots__ = ["chat", "msgid"]

    def __init__(self, chat: int, msgid: int) -> None:
        self.chat = chat
        self.msgid = msgid

    def __hash__(self) -> int:
        return hash((self.chat, self.msgid))

    def __eq__(self, o) -> bool:
        if not isinstance(o, ChatMessage):
            return False
        try:
            return self.chat == o.chat and self.msgid == o.msgid
        except Exception:
            return False

    def __expr__(self) -> str:
        return str(self.chat) + ' ' + str(self.msgid)

    def __str__(self) -> str:
        return self.__expr__()


def get_from_id(update: Update) -> Optional[int]:
    """返回`from_user.id`"""
    if update.message is not None and update.message.from_user is not None:
        return update.message.from_user.id
    if update.callback_query is not None and update.callback_query.from_user is not None:
        return update.callback_query.from_user.id
    if update.channel_post is not None and update.channel_post.from_user is not None:
        return update.channel_post.from_user.id
    if update.edited_channel_post is not None and update.edited_channel_post.from_user is not None:
        return update.edited_channel_post.from_user.id
    return None


def get_chat_id(update: Update) -> Optional[int]:
    """返回`chat_id`"""
    if update.effective_chat is not None:
        return update.effective_chat.id
    return None


def get_msg_id(update: Update) -> Optional[int]:
    """返回message_id"""
    if update.message is not None:
        return update.message.message_id
    if update.callback_query is not None:
        return update.callback_query.message.message_id  # type: ignore
    if update.channel_post is not None:
        return update.channel_post.message_id
    if update.edited_channel_post is not None:
        return update.edited_channel_post.message_id
    return None


def get_reply_to_msg_id(update: Update) -> Optional[int]:
    if update.message is not None and update.message.reply_to_message is not None:
        return update.message.reply_to_message.message_id
    if update.callback_query is not None and update.callback_query.message.reply_to_message is not None:  # type: ignore
        return update.callback_query.message.reply_to_message.message_id  # type: ignore
    if update.channel_post is not None and update.channel_post.reply_to_message is not None:
        return update.channel_post.reply_to_message.message_id
    if update.edited_channel_post is not None and update.edited_channel_post.reply_to_message is not None:
        return update.edited_channel_post.reply_to_message.message_id
    return None


def is_private(update: Update) -> bool:
    return update.effective_chat.type == ChatType.PRIVATE  # type: ignore


def is_group(update: Update) -> bool:
    return update.effective_chat.type in [ChatType.GROUP, ChatType.SUPERGROUP]  # type: ignore


def is_channel(update: Update) -> bool:
    return update.effective_chat.type == ChatType.CHANNEL  # type: ignore


def flatten_button(
    buttons: List[InlineKeyboardButton], numberInOneLine: int
) -> InlineKeyboardMarkup:
    btl: List[List[InlineKeyboardButton]] = []
    while len(buttons) > numberInOneLine:
        btl.append(buttons[:numberInOneLine])
        buttons = buttons[numberInOneLine:]
    if len(buttons) > 0:
        btl.append(buttons)
    return InlineKeyboardMarkup(btl)


async def exception_manual_handle(logger: "logging.Logger", e: Exception):
    """
    no raise
    """
    try:
        logger.debug("exception catched, manually handling it")
        from antares_bot.bot_inst import exception_handler
        from antares_bot.context_manager import get_context
        context = get_context()
        context.error = e
    except Exception as _e:
        try:
            from antares_bot.bot_inst import format_traceback
            logger.error("when manually handling exception, another exception raised: %s", format_traceback(_e))
        except Exception:
            pass
        return
    # exception_handler does not raise
    await exception_handler(None, context)


async def fetch_url(url: str, **kwargs):
    async with httpx.AsyncClient() as client:
        return await client.get(url, **kwargs)


def markdown_v2_escape(s: str) -> str:
    """
    Escape markdown special characters.
    Reference: https://core.telegram.org/bots/api#markdownv2-style
    """
    # note that ~ < > in telegram is also special
    special_chars = ['_', '*', '[', ']', '(', ')', '~', '`', '>', '#', '+', '-', '=', '|', '{', '}', '.', '!']
    for c in special_chars:
        s = s.replace(c, "\\" + c)
    return s


def markdown_escape(s: str) -> str:
    """
    Escape markdown special characters.
    Reference: https://core.telegram.org/bots/api#markdown-style
    """
    special_chars = ['`', '*', '_', '[']
    for c in special_chars:
        s = s.replace(c, "\\" + c)
    return s


def systemd_service_info():
    name = read_user_cfg(AntaresBotConfig, "SYSTEMD_SERVICE_NAME")
    if name is not None:
        is_root = os.geteuid() == 0
        return name, is_root
    return None, None


def merge_dicts(dict1: dict, dict2: dict):
    merged = dict1.copy()
    for key, value2 in dict2.items():
        value1 = merged.get(key)
        if isinstance(value1, dict) and isinstance(value2, dict):
            merged[key] = merge_dicts(value1, value2)
        else:
            merged[key] = value2
    return merged



import importlib
import sys
from types import ModuleType
from typing import Optional, Type, TypeVar, cast

from antares_bot.bot_logging import get_logger


_T = TypeVar("_T")

_LOGGER = get_logger(__name__)


def get_module_from_name(module_name: str):
    """
    Return:
        module, is_reload

    Note:
        Does not raise any exception
    """
    try:
        if module_name in sys.modules:
            return sys.modules[module_name], True
        else:
            return importlib.import_module(module_name), False
    except Exception as e:
        _LOGGER.error(e)
        return None, False


def get_module_class_from_module(module: ModuleType, instance_class: Optional[Type[_T]] = None) -> Optional[Type[_T]]:
    """
    Return:
        kls

    Note:
        Does not reload. Does not raise any exception
    """
    try:
        modulefile_name = module.__name__.split(".")[-1]
        modulefile_names = modulefile_name.split("_")
        class_name = ''.join([name.capitalize() for name in modulefile_names])
        kls = getattr(module, class_name, None)
        if instance_class is None:
            return kls
        # need check class
        if issubclass(cast(type, kls), instance_class):
            return kls
    except Exception:
        pass
    return None


def get_module_class_from_name(module_name: str, instance_class: Optional[Type[_T]] = None) -> Optional[Type[_T]]:
    """
    Return:
        kls (optional)

    Note:
        Does not reload. Does not raise any exception
    """
    module, _ = get_module_from_name(module_name)
    if module is None:
        return None
    return get_module_class_from_module(module, instance_class)



import asyncio
from typing import TYPE_CHECKING, Any, AsyncGenerator, Awaitable, Callable, List, TypeVar

from telegram.error import BadRequest, ChatMigrated, Forbidden, InvalidToken, RetryAfter, TelegramError

from antares_bot.bot_logging import get_logger
from antares_bot.text_process import longtext_split


if TYPE_CHECKING:
    from telegram import Message

    from antares_bot.context import RichCallbackContext


_T = TypeVar("_T")
_LOGGER = get_logger("bot_base")


class TelegramBotBaseWrapper(object):
    RETRY_TIMES = 3
    RETRY_SLEEP_TIME = 1.

    @classmethod
    def get_context(cls) -> "RichCallbackContext":
        raise NotImplementedError

    @staticmethod
    async def _get_last(gen: AsyncGenerator[_T, Any]) -> _T:
        last = None
        async for m in gen:
            last = m
        assert last is not None
        return last

    @classmethod
    async def success_info(cls, text: str, **kwargs):
        await cls.reply(text, **kwargs)
        return True

    @classmethod
    async def error_info(cls, text: str, **kwargs):
        await cls.reply(text, **kwargs)
        return False

    ##############################

    @classmethod
    async def reply_to(cls, message: "Message", text: str, **kwargs) -> int:
        return (await cls._get_last(cls._reply_to(message, text, **kwargs))).id

    @classmethod
    async def reply_to_v2(cls, message: "Message", text: str, **kwargs) -> List[int]:
        return [m.id async for m in cls._reply_to(message, text, **kwargs)]

    @classmethod
    async def reply_to_v3(cls, message: "Message", text: str, **kwargs) -> "Message":
        return await cls._get_last(cls._reply_to(message, text, **kwargs))

    @classmethod
    async def reply_to_v4(cls, message: "Message", text: str, **kwargs) -> List["Message"]:
        return [m async for m in cls._reply_to(message, text, **kwargs)]

    @classmethod
    async def _reply_to(cls, message: "Message", text: str, **kwargs):
        if "entities" in kwargs:
            texts = [text]
        else:
            texts = longtext_split(text)
        if 'reply_to_message_id' not in kwargs:
            kwargs['reply_to_message_id'] = message.id
        async for m in cls._sequence_send(message.reply_text, texts, **kwargs):
            yield m

    ##############################

    @classmethod
    async def reply(cls, text: str, **kwargs):
        return (await cls._get_last(cls._reply(text, **kwargs))).id

    @classmethod
    async def reply_v2(cls, text: str, **kwargs):
        return [m.id async for m in cls._reply(text, **kwargs)]

    @classmethod
    async def reply_v3(cls, text: str, **kwargs):
        return await cls._get_last(cls._reply(text, **kwargs))

    @classmethod
    async def reply_v4(cls, text: str, **kwargs):
        return [m async for m in cls._reply(text, **kwargs)]

    @classmethod
    async def _reply(cls, text: str, **kwargs):
        context = cls.get_context()
        kwargs.setdefault('chat_id', context.chat_id)
        chat_id: int = kwargs['chat_id']
        if chat_id == context.chat_id and context.message_id is not None and 'reply_to_message_id' not in kwargs and not context.is_callback_query():
            kwargs['reply_to_message_id'] = context.message_id
        if "entities" in kwargs:
            texts = [text]
        else:
            texts = longtext_split(text)
        async for m in cls._sequence_send(context.bot.send_message, texts, **kwargs):
            yield m

    ##############################

    @classmethod
    async def send_to(cls, chat_id: int, text: str, **kwargs):
        return (await cls._get_last(cls._send_to(chat_id, text, **kwargs))).id

    @classmethod
    async def send_to_v2(cls, chat_id: int, text: str, **kwargs):
        return [m.id async for m in cls._send_to(chat_id, text, **kwargs)]

    @classmethod
    async def send_to_v3(cls, chat_id: int, text: str, **kwargs):
        return await cls._get_last(cls._send_to(chat_id, text, **kwargs))

    @classmethod
    async def send_to_v4(cls, chat_id: int, text: str, **kwargs):
        return [m async for m in cls._send_to(chat_id, text, **kwargs)]

    @classmethod
    async def _send_to(cls, chat_id: int, text: str, **kwargs):
        from antares_bot.bot_inst import get_bot_instance
        if "entities" in kwargs:
            texts = [text]
        else:
            texts = longtext_split(text)
        kwargs['chat_id'] = chat_id
        async for m in cls._sequence_send(get_bot_instance().bot.send_message, texts, **kwargs):
            yield m

    ##############################

    @classmethod
    async def send_photo(cls, chat_id, photo, **kwargs) -> "Message":
        kwargs['chat_id'] = chat_id
        kwargs['photo'] = photo
        from antares_bot.bot_inst import get_bot_instance
        return await cls._send_ignore_parsemode_or_replyto_exceptions(get_bot_instance().bot.send_photo, _no_retry=True, **kwargs)

    @classmethod
    async def reply_photo(cls, photo, **kwargs) -> "Message":
        context = cls.get_context()
        kwargs.setdefault('chat_id', context.chat_id)
        chat_id: int = kwargs['chat_id']
        if chat_id == context.chat_id and context.message_id is not None and 'reply_to_message_id' not in kwargs and not context.is_callback_query():
            kwargs['reply_to_message_id'] = context.message_id
        kwargs['photo'] = photo
        return await cls._send_ignore_parsemode_or_replyto_exceptions(context.bot.send_photo, _no_retry=True, **kwargs)

    @classmethod
    async def send_document(cls, chat_id, document, **kwargs) -> "Message":
        kwargs['chat_id'] = chat_id
        kwargs['document'] = document
        from antares_bot.bot_inst import get_bot_instance
        return await cls._send_ignore_parsemode_or_replyto_exceptions(get_bot_instance().bot.send_document, _no_retry=True, **kwargs)

    @classmethod
    async def reply_document(cls, document, **kwargs) -> "Message":
        context = cls.get_context()
        kwargs.setdefault('chat_id', context.chat_id)
        chat_id: int = kwargs['chat_id']
        if chat_id == context.chat_id and context.message_id is not None and 'reply_to_message_id' not in kwargs and not context.is_callback_query():
            kwargs['reply_to_message_id'] = context.message_id
        kwargs['document'] = document
        return await cls._send_ignore_parsemode_or_replyto_exceptions(context.bot.send_document, _no_retry=True, **kwargs)

    ##############################

    @classmethod
    async def _sequence_send(cls, interface_func: Callable[..., Awaitable["Message"]], texts: List[str], **kwargs):
        rp_markup = kwargs.pop('reply_markup', None)
        for i, t in enumerate(texts):
            if rp_markup is not None and i == len(texts) - 1:
                kwargs['reply_markup'] = rp_markup
            yield await cls._send_ignore_parsemode_or_replyto_exceptions(interface_func, text=t, **kwargs)

    @classmethod
    async def _retry_call(cls, func: Callable[..., Awaitable[_T]], *args, **kwargs):
        # not using `raise from` to avoid long traceback
        for i in range(cls.RETRY_TIMES):
            try:
                return await func(*args, **kwargs)
            except (BadRequest, ChatMigrated, Forbidden, InvalidToken) as e:
                raise e  # exit fast
            except TelegramError as e:
                if i == cls.RETRY_TIMES - 1:
                    raise e
                try:
                    _LOGGER.warning("retrying %s %d/%d due to %s", func.__name__, i + 1, cls.RETRY_TIMES, str(e))
                except Exception:
                    pass
                _sleep_time = e.retry_after + 1 if isinstance(e, RetryAfter) else cls.RETRY_SLEEP_TIME
                await asyncio.sleep(_sleep_time)
        raise RuntimeError(f"unreachable: RETRY_TIMES={cls.RETRY_TIMES}")

    @classmethod
    async def _send_ignore_parsemode_or_replyto_exceptions(
        cls,
        interface_func: Callable[..., Awaitable["Message"]],
        _no_retry=False,
        **kwargs
    ) -> "Message":
        try:
            if _no_retry:
                return await interface_func(**kwargs)
            return await cls._retry_call(interface_func, **kwargs)
        except BadRequest as e:
            if str(e).find("reply") != -1:
                _LOGGER.error("send message failed, retrying by popping reply_to_message_id")
                if kwargs.pop('reply_to_message_id', None) is not None:
                    return await cls._send_ignore_parsemode_or_replyto_exceptions(interface_func, _no_retry, **kwargs)
            if str(e).find('parse') != -1:
                _LOGGER.error("send message failed, retrying by popping parse_mode")
                parse_mode = kwargs.pop('parse_mode', None)
                if parse_mode is None:
                    raise
                ret = await cls._send_ignore_parsemode_or_replyto_exceptions(interface_func, _no_retry, **kwargs)
                # reset parse_mode for the next call
                kwargs['parse_mode'] = parse_mode
                return ret
            raise



import weakref
from typing import TYPE_CHECKING, Any, Optional

from telegram.ext import JobQueue

from antares_bot.bot_default_cfg import AntaresBotConfig
from antares_bot.context_manager import ContextReverseHelper
from antares_bot.utils import merge_dicts, read_user_cfg


try:
    import pytz  # type: ignore
    from apscheduler.executors.asyncio import AsyncIOExecutor  # type: ignore
    from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore

    APS_AVAILABLE = True
except ImportError:
    APS_AVAILABLE = False

if TYPE_CHECKING:
    from telegram.ext import Application


_MISFIRE_GRACE_TIME = 60  # seconds


class _PatchedAsyncIOScheduler(AsyncIOScheduler):
    def add_job(self, *args, **kwargs):
        with ContextReverseHelper():
            return super().add_job(*args, **kwargs)


class JobQueueEx(JobQueue):
    def __init__(self) -> None:  # pylint: disable=super-init-not-called
        if not APS_AVAILABLE:
            raise RuntimeError(
                "To use `JobQueueEx`, PTB must be installed via `pip install "
                '"python-telegram-bot[job-queue]"`.'
            )

        self._application: Optional[weakref.ReferenceType["Application"]] = None
        self._executor = AsyncIOExecutor()
        self.scheduler: AsyncIOScheduler = _PatchedAsyncIOScheduler(
            timezone=pytz.utc, executors={"default": self._executor}
        )

    @property
    def scheduler_configuration(self):
        ret = super().scheduler_configuration
        scheduler_config = read_user_cfg(AntaresBotConfig, "JOB_QUEUE_CONFIG")
        if scheduler_config is not None:
            ret = merge_dicts(ret, scheduler_config)
        fix_misfire_grace_time(ret)
        return ret


def fix_misfire_grace_time(d: dict[str, Any]):
    d.setdefault("job_defaults", {}).setdefault("misfire_grace_time", _MISFIRE_GRACE_TIME)



from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Coroutine, TypeVar, overload

from antares_bot.context import RichCallbackContext


if TYPE_CHECKING:
    from contextvars import Token


_JobCallback = Callable[[RichCallbackContext], Awaitable[Any]]
_CoroutineJobCallback = Callable[[RichCallbackContext], Coroutine[Any, Any, Any]]
_T = TypeVar("_T", _JobCallback, _CoroutineJobCallback)

context_manager: ContextVar[RichCallbackContext] = ContextVar('RichCallbackContext', default=None)  # type: ignore


def get_context():
    return context_manager.get()


def set_context(context: RichCallbackContext):
    return context_manager.set(context)


def reset_context(token: "Token[RichCallbackContext]"):
    context_manager.reset(token)


@overload
def callback_job_wrapper(context: RichCallbackContext) -> Callable[[_T], _T]:
    ...


@overload
def callback_job_wrapper(func: _T) -> _T:
    ...


def callback_job_wrapper(arg):
    if isinstance(arg, RichCallbackContext):
        def wrapper(func):
            async def wrapped(*args, **kwargs):
                with ContextHelper(arg):
                    return await func(*args, **kwargs)
            return wrapped
        return wrapper

    # is function
    _ct = get_context()

    async def wrapped(*args, **kwargs):
        with ContextHelper(_ct):
            return await arg(*args, **kwargs)
    return wrapped


class ContextHelper:
    __slots__ = ("context", "token")

    def __init__(self, context: RichCallbackContext):
        self.context = context
        self.token = None

    def __enter__(self):
        self.token = set_context(self.context)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        reset_context(self.token)
        self.token = None
        return False


class InvalidContextError(RuntimeError):
    pass


class InvalidContext:
    def __getattr__(self, _):
        raise InvalidContextError("Invalid context. Did you forget to use `callback_job_wrapper` when creating callback?")


class ContextReverseHelper:
    def __init__(self):
        self.token = None

    def __enter__(self):
        self.token = context_manager.set(InvalidContext())
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        context_manager.reset(self.token)
        self.token = None
        return False



from typing import Any, Generator, Literal, Optional, Sequence, Union

from telegram import Document, Message, MessageEntity, PhotoSize
from telegram._utils.defaultvalue import DEFAULT_NONE
from telegram._utils.types import DVInput, FileInput, JSONDict, ODVInput, ReplyMarkup
from telegram.ext._utils.types import RLARGS


class TelegramBotBaseWrapper(object):
    @classmethod
    async def success_info(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Literal[True]:
        ...

    @classmethod
    async def error_info(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Literal[False]:
        ...

    @classmethod
    async def reply_to(
        cls,
        message: Message,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        quote: Optional[bool] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
    ) -> int:
        ...

    @classmethod
    async def reply_to_v2(
        cls,
        message: Message,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        quote: Optional[bool] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
    ) -> list[int]:
        ...

    @classmethod
    async def reply_to_v3(
        cls,
        message: Message,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        quote: Optional[bool] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
    ) -> Message:
        ...

    @classmethod
    async def reply_to_v4(
        cls,
        message: Message,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        quote: Optional[bool] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
    ) -> list[Message]:
        ...

    @classmethod
    async def _reply_to(
        cls,
        message: Message,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        quote: Optional[bool] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
    ) -> Generator[Message, Any, None]:
        ...

    @classmethod
    async def reply(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> int:
        ...

    @classmethod
    async def reply_v2(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> list[int]:
        ...

    @classmethod
    async def reply_v3(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Message:
        ...

    @classmethod
    async def reply_v4(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> list[Message]:
        ...

    @classmethod
    async def _reply(
        cls,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Generator[Message, Any, None]:
        ...

    @classmethod
    async def send_to(
        cls,
        chat_id: int,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> int:
        ...

    @classmethod
    async def send_to_v2(
        cls,
        chat_id: int,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> list[int]:
        ...

    @classmethod
    async def send_to_v3(
        cls,
        chat_id: int,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Message:
        ...

    @classmethod
    async def send_to_v4(
        cls,
        chat_id: int,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> list[Message]:
        ...

    @classmethod
    async def _send_to(
        cls,
        chat_id: int,
        text: str,
        *,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        entities: Optional[Sequence[MessageEntity]] = None,
        disable_web_page_preview: ODVInput[bool] = DEFAULT_NONE,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        reply_markup: Optional[ReplyMarkup] = None,
        message_thread_id: Optional[int] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = DEFAULT_NONE,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Generator[Message, Any, None]:
        ...
    ##############################

    @classmethod
    async def send_photo(
        cls,
        chat_id: Union[int, str],
        photo: Union[FileInput, PhotoSize],
        *,
        caption: Optional[str] = None,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        caption_entities: Optional[Sequence["MessageEntity"]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        has_spoiler: Optional[bool] = None,
        filename: Optional[str] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = 20,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None
    ) -> Message:
        ...

    @classmethod
    async def reply_photo(
        cls,
        photo: Union[FileInput, PhotoSize],
        *,
        chat_id: Union[int, str, None] = None,
        caption: Optional[str] = None,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        caption_entities: Optional[Sequence["MessageEntity"]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        has_spoiler: Optional[bool] = None,
        filename: Optional[str] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = 20,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None
    ) -> Message:
        ...

    @classmethod
    async def send_document(
        cls,
        chat_id: Union[int, str],
        document: Union[FileInput, Document],
        *,
        caption: Optional[str] = None,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_content_type_detection: Optional[bool] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        caption_entities: Optional[Sequence["MessageEntity"]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        thumbnail: Optional[FileInput] = None,
        filename: Optional[str] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = 20,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Message:
        ...

    @classmethod
    async def reply_document(
        cls,
        document: Union[FileInput, Document],
        *,
        chat_id: Union[int, str, None] = None,
        caption: Optional[str] = None,
        disable_notification: DVInput[bool] = DEFAULT_NONE,
        reply_to_message_id: Optional[int] = None,
        reply_markup: Optional[ReplyMarkup] = None,
        parse_mode: ODVInput[str] = DEFAULT_NONE,
        disable_content_type_detection: Optional[bool] = None,
        allow_sending_without_reply: ODVInput[bool] = DEFAULT_NONE,
        caption_entities: Optional[Sequence["MessageEntity"]] = None,
        protect_content: ODVInput[bool] = DEFAULT_NONE,
        message_thread_id: Optional[int] = None,
        thumbnail: Optional[FileInput] = None,
        filename: Optional[str] = None,
        read_timeout: ODVInput[float] = DEFAULT_NONE,
        write_timeout: ODVInput[float] = 20,
        connect_timeout: ODVInput[float] = DEFAULT_NONE,
        pool_timeout: ODVInput[float] = DEFAULT_NONE,
        api_kwargs: Optional[JSONDict] = None,
        rate_limit_args: Optional[RLARGS] = None,
    ) -> Message:
        ...



import logging
from logging import Handler, getLevelName
from types import GenericAlias
from typing import Optional, cast

from antares_bot.bot_default_cfg import BasicConfig
from antares_bot.init_hooks import read_user_cfg


try:
    from pika_interface import close_sustained_connection, send_message_nowait

    PIKA_SUPPORTED = True
    __pika_logger_stopped = False
except ImportError:
    print(
        "Pika not supported due to an ImportError."
        " Safely ignore it if you do not use pika."
        " Check whether aio-pika is installed if you use pika."
    )
    PIKA_SUPPORTED = False

_logger_top_name = "antares_bot"
__logger_inited = False
__root_logger = cast(logging.Logger, None)


if PIKA_SUPPORTED:
    import threading

    def _is_pika_logger_running():
        if not (threading.current_thread() is threading.main_thread()):
            return False
        return not __pika_logger_stopped

    class PikaHandler(Handler):
        def emit(self, record):
            if _is_pika_logger_running():
                try:
                    msg = self.format(record)
                    key = record.name
                    if not key.startswith(_logger_top_name):
                        key = _logger_top_name + "." + key
                    send_message_nowait("logging." + key, msg)
                except RecursionError:  # See issue 36272
                    raise
                except Exception:
                    self.handleError(record)

        def __repr__(self):
            level = getLevelName(self.level)
            name = self.name
            #  bpo-36015: name can be an int
            name = str(name)
            if name:
                name += ' '
            return '<%s %s(%s)>' % (self.__class__.__name__, name, level)

        __class_getitem__ = classmethod(GenericAlias)  # type: ignore


def _log_start(logger_top_name: Optional[str] = None) -> logging.Logger:
    global __logger_inited, __root_logger, _logger_top_name
    if __logger_inited:
        return __root_logger
    __logger_inited = True
    if logger_top_name is not None:
        _logger_top_name = logger_top_name

    __root_logger = logging.getLogger(_logger_top_name)
    if PIKA_SUPPORTED:
        handler = PikaHandler()
        __root_logger.addHandler(handler)
        # add handler to telegram internal logger
        tg_logger = logging.getLogger("telegram")
        tg_logger.addHandler(handler)
        # add handler to apscheduler internal logger
        apscheduler_logger = logging.getLogger("apscheduler")
        apscheduler_logger.addHandler(handler)
    return __root_logger


def get_logger(module_name: str):
    strip_prefix = "modules."
    if module_name.startswith(strip_prefix):
        module_name = module_name[len(strip_prefix):]
    if not __logger_inited:
        raise RuntimeError("logger not inited")
    name = _logger_top_name + "." + module_name
    logger = logging.getLogger(name)
    return logger


def add_pika_log_handler(logger: str | logging.Logger) -> bool:
    """
    Add pika handler for the logger.
    One should only add pika handler to the logger after the logger is initialized.
    The handler will take effect for the logger and all its children.
    """
    if not PIKA_SUPPORTED:
        print("Pika not supported, add_pika_log_handler has no effect.")
        return False
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    added = False
    for handler in __root_logger.handlers:
        if isinstance(handler, PikaHandler):
            logger.addHandler(handler)
            added = True
            break
    return added


def get_root_logger():
    if not __logger_inited:
        raise RuntimeError("logger not inited")
    return __root_logger


async def stop_logger():
    if PIKA_SUPPORTED:
        global __pika_logger_stopped
        __pika_logger_stopped = True
        await close_sustained_connection()


_log_start(read_user_cfg(BasicConfig, "BOT_NAME"))



class BaseConfig:
    @classmethod
    def iter_all_config_keys(cls):
        keys = dir(cls)
        for k in keys:
            if k.startswith("__") or k != k.upper():
                continue
            yield k


class BasicConfig(BaseConfig):
    non_empty = [
        "TOKEN",
        "MASTER_ID",
    ]
    """
    Basic config.
    Create a class with same name in bot_cfg.py to override the default value.
    TOKEN and MASTER_ID must be overriden.
    """
    LOCALE = 'en'
    DATA_DIR = "data"
    TOKEN = "abcdef:123456"
    MASTER_ID = 123456789


class AntaresBotConfig(BaseConfig):
    """
    Config to control the bot behavior.
    Create a class with same name in bot_cfg.py to set the value.
    """
    # SKIP_PIKA_SETUP = True
    # SKIP_LOAD_ALL_INTERNAL_MODULES = True
    # SKIP_LOAD_ALL_MODULES = True
    # SKIP_LOAD_INTERNAL_MODULE_{module_name_upper_case} = True
    # SKIP_LOAD_MODULE_{module_name_upper_case} = True
    # OBJGRAPH_TRACE_AT_START = True



import asyncio
import datetime
import os
import signal
import subprocess
import sys
import time
import traceback
from logging import DEBUG as LOGLEVEL_DEBUG
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast

from telegram import MessageOriginChannel, MessageOriginChat, MessageOriginUser, Update
from telegram.error import Conflict, NetworkError, RetryAfter, TimedOut
from telegram.ext import Application, CommandHandler, ContextTypes, ConversationHandler

from antares_bot.basic_language import BasicLanguage as Lang
from antares_bot.bot_base import TelegramBotBase
from antares_bot.bot_default_cfg import AntaresBotConfig, BasicConfig
from antares_bot.bot_logging import get_logger, get_root_logger, stop_logger
from antares_bot.callback_manager import CallbackDataManager
from antares_bot.context import ChatData, RichCallbackContext, UserData
from antares_bot.context_manager import ContextHelper, ContextReverseHelper, get_context
from antares_bot.error import InvalidChatTypeException, UserPermissionException, permission_exceptions
from antares_bot.framework import CallbackBase, command_callback_wrapper
from antares_bot.module_loader import ModuleKeeper
from antares_bot.patching.job_quque_ex import JobQueueEx
from antares_bot.permission_check import CheckLevel
from antares_bot.sqlite.manager import DataBasesManager
from antares_bot.text_process import trim_spaces_before_line
from antares_bot.utils import SYSTEM_TIME_ZONE, markdown_escape, read_user_cfg, systemd_service_info


if TYPE_CHECKING:
    from telegram.ext import ExtBot

    from antares_bot.module_base import TelegramBotModuleBase

_T = TypeVar("_T", bound="TelegramBotModuleBase", covariant=True)

_LOGGER = get_logger("main")
TIME_IN_A_DAY = 24 * 60 * 60

_PROGRAM_SHUTDOWN_STARTED = False

# note that wildcard import is only allowed at module level
_INTERNAL_TEST_EXEC_COMMAND_PREFIX = """\
from antares_bot.test_commands import *
from bot_cfg import BasicConfig
MASTER_ID = BasicConfig.MASTER_ID
async def __t():
    self = get_bot_instance()
"""


def format_traceback(err: Exception) -> str:
    return '\n'.join(traceback.format_tb(err.__traceback__))


async def _leaked_permission_exception_handler(err: Exception):
    try:
        if isinstance(err, UserPermissionException):
            await get_bot_instance().reply(Lang.t(Lang.NO_PERMISSION))
        elif isinstance(err, InvalidChatTypeException):
            await get_bot_instance().reply(Lang.t(Lang.INVALID_CHAT_TYPE))
    except Exception:
        _LOGGER.error(format_traceback(err))


async def exception_handler(update: Any, context: RichCallbackContext):
    try:
        err = context.error
        if err is None or err.__class__ in (NetworkError, OSError, TimedOut, ConnectionError, Conflict, RetryAfter):
            return  # ignore them
        permission_exception_types = permission_exceptions()
        if isinstance(err, permission_exception_types):
            # in case of didn't catching permission exceptions properly
            # generally, catching permission exception here greatly affects the performance
            if get_context() is None:
                _LOGGER.error("No context found")
                return
            with ContextHelper(context):
                await _leaked_permission_exception_handler(err)
            return
        tb = format_traceback(err)
        log_text = f"{err.__class__}\n{err}\ntraceback:\n{tb}"
        _LOGGER.error(log_text)
        text = Lang.t(Lang.UNKNOWN_ERROR) + f"\n{err.__class__}: {err}"
        await get_bot_instance().send_to(TelegramBot.get_master_id(), text)
    except Exception as _e:
        try:
            _LOGGER.error(format_traceback(_e))
        except Exception:
            pass


class TelegramBot(TelegramBotBase):
    ContextType = RichCallbackContext
    ChatDataType = ChatData
    UserDataType = UserData
    BotDataType = dict

    def __init__(self) -> None:
        self._module_keeper = ModuleKeeper()
        context_types = ContextTypes(
            context=self.ContextType,
            chat_data=self.ChatDataType,
            user_data=self.UserDataType,
            bot_data=self.BotDataType  # type: ignore
        )
        from telegram.ext import Defaults
        defaults = Defaults(tzinfo=SYSTEM_TIME_ZONE)
        self.application = cast(
            "Application[ExtBot[None], self.ContextType, self.UserDataType, self.ChatDataType, self.BotDataType, JobQueueEx]",
            Application.builder()
            # .application_class(ApplicationEx)
            .token(read_user_cfg(BasicConfig, "TOKEN"))
            .context_types(context_types)
            .job_queue(JobQueueEx())
            .post_init(self._do_post_init)
            .post_stop(self._do_post_stop)
            .defaults(defaults)
            .build()
        )
        self.bot = self.application.bot
        assert self.application.updater is not None
        self.updater = self.application.updater
        assert self.application.job_queue is not None
        self.job_queue = self.application.job_queue
        #
        self.callback_manager = CallbackDataManager()
        self.callback_key_dict: Dict[Tuple[int, int], List[str]] = dict()
        self._old_log_level = None
        self._custom_post_init_task: Awaitable | None = None
        self._custom_post_stop_task: Awaitable | None = None
        # TODO do a flags check at the end of the run. move the flags into a new class
        self._post_stop_restart_flag = False
        self._post_stop_gitpull_flag = bool(read_user_cfg(AntaresBotConfig, "PULL_WHEN_STOP"))
        self._custom_restart_command: str | list[str] | None = None
        self._custom_finalize_task: Callable[[], Any] | None = None
        self._normal_exit_flag = False
        self.handler_docs: dict[str, str] = {}
        self._exit_fast = False
        # some pre-checks
        if self._is_debug_level():
            _LOGGER.debug("Warning: the initial logging level is DEBUG. The built-in /debug_mode command will not work.")

    @property
    def bot_id(self):
        return self.bot.bot.id

    async def _do_post_init(self, app: Application):
        time0 = time.time()
        tasks: list[Awaitable] = [module.post_init(app) for module in self._module_keeper.get_all_enabled_modules()]
        if self._custom_post_init_task is not None:
            tasks.append(self._custom_post_init_task)
            self._custom_post_init_task = None
        tasks.append(self.send_init_hello())
        await asyncio.gather(*tasks)
        time1 = time.time()
        _LOGGER.warning("Post init time: %.3fs", time1 - time0)

    async def send_init_hello(self) -> None:
        await self.send_to(self.get_master_id(), Lang.t(Lang.STARTUP))

    async def _do_post_stop(self, app: Application):
        time0 = time.time()
        if self._custom_post_stop_task is not None:
            await self._custom_post_stop_task
            self._custom_post_stop_task = None
        task_send_exit_msg = self.send_to(self.get_master_id(), Lang.t(Lang.SHUTDOWN_GOODBYTE))
        await asyncio.gather(*(module.do_stop() for module in self._module_keeper.get_all_enabled_modules()))
        time1 = time.time()
        _LOGGER.warning("Post stop time: %.3fs", time1 - time0)
        task_stop_db = DataBasesManager.get_inst().shutdown()
        time2 = time.time()
        _LOGGER.warning("Shutdown db time: %.3fs", time2 - time1)
        # pull the repo if _post_stop_gitpull_flag is set.
        # if exit_fast (SIGTERM, SIGABRT), do not pull
        additional_tasks = []
        if self._post_stop_gitpull_flag and not self._exit_fast:
            # create a subprocess to git pull
            try:
                msg = str(subprocess.check_output(
                    ["git", "pull", "--ff-only"],
                    encoding='utf-8')
                )
                if msg and "Already up to date." not in msg and "not a git repository" not in msg:
                    additional_tasks.append(self.send_to(self.get_master_id(), msg))
            except Exception:
                additional_tasks.append(self.send_to(self.get_master_id(), "git pull failed!"))
        task_stop_logger = stop_logger()
        # wait for all tasks to finish
        await asyncio.gather(task_send_exit_msg, task_stop_db, task_stop_logger, *additional_tasks)

    def custom_post_init(self, task: Awaitable):
        self._custom_post_init_task = task

    def custom_post_stop(self, task: Awaitable):
        self._custom_post_stop_task = task

    def custom_restart_command(self, command: str | list[str]):
        self._custom_restart_command = command

    def custom_finalize(self, finalize_task: Callable[[], Any]):
        self._custom_finalize_task = finalize_task

    def run(self):
        self._module_keeper.load_all()
        for module in self._module_keeper.get_all_enabled_modules():
            module.do_init(self)

        for module in self._module_keeper.get_all_enabled_modules():
            module_inst = module.module_instance
            for func in module_inst.collect_handlers():
                if isinstance(func, CallbackBase):
                    handler = func.to_handler()
                else:
                    handler = func
                if isinstance(handler, CommandHandler):
                    for command in handler.commands:
                        _doc = func.__doc__
                        self.handler_docs[command] = _doc if _doc else "No doc"
                elif isinstance(handler, ConversationHandler):
                    entry = handler.entry_points
                    for entry_point in entry:
                        if isinstance(entry_point, CommandHandler):
                            for command in entry_point.commands:
                                _doc = entry_point.callback.__doc__
                                self.handler_docs[command] = _doc if _doc else "No doc"
                self.application.add_handler(handler)
                # try get module logger
                py_module = module.py_module()
                if hasattr(py_module, "_LOGGER"):
                    logger = getattr(py_module, "_LOGGER")
                else:
                    logger = get_logger(module.top_name)
                    setattr(py_module, "_LOGGER", logger)
                logger.info("added handler: %s", func)

        main_handlers = [
            self.stop,
            self.restart,
            self.debug_mode,
            self.exec,
            self.get_id,
            self.help,
        ]

        for method in main_handlers:
            handler = method.to_handler()  # pylint: disable=no-member
            self.application.add_handler(handler)
            for command in handler.commands:
                self.handler_docs[command] = method.__doc__ if method.__doc__ else "No doc"
            _LOGGER.info("added handler: %s", handler)

        self.handler_docs["cancel"] = """
        cancel - cancel the current operation
        `/cancel`: Cancel the current operation.
        """

        self.application.add_error_handler(exception_handler)

        self.job_queue.run_daily(self._daily_job, time=datetime.time(hour=0, minute=0, tzinfo=SYSTEM_TIME_ZONE), name="daily_job")

        signal.signal(signal.SIGINT, self.signal_stop)
        signal.signal(signal.SIGTERM, self.signal_stop)
        signal.signal(signal.SIGABRT, self.signal_stop)

        # use eager factory for python 3.12+
        if sys.version_info >= (3, 12):
            asyncio.get_event_loop().set_task_factory(asyncio.eager_task_factory)

        try:
            self.application.run_polling(
                allowed_updates=Update.ALL_TYPES,
                drop_pending_updates=True,
                stop_signals=(),
            )
        except NetworkError:
            # catches the NetworkError when the bot is turned off.
            # we don't care about that when normal exit
            if not self._normal_exit_flag:
                raise

        # post run
        self._post_run()

    def _post_run(self):
        global _PROGRAM_SHUTDOWN_STARTED
        if self._custom_finalize_task is not None:
            print("Running custom finalize task...")
            self._custom_finalize_task()

        if self._post_stop_restart_flag and not _PROGRAM_SHUTDOWN_STARTED:
            _PROGRAM_SHUTDOWN_STARTED = True
            systemd_service_name, is_root = systemd_service_info()
            if systemd_service_name is not None:
                restart_command = ["systemctl"] + (["--user"] if not is_root else []) + ["restart", systemd_service_name]
                restart_command = ' '.join(restart_command)
            else:
                # create a detached subprocess to restart the bot
                restart_command = self._custom_restart_command if self._custom_restart_command is not None else sys.orig_argv
                if isinstance(restart_command, list):
                    import shlex
                    restart_command = shlex.join(restart_command)
                restart_command = f"{restart_command} & disown"
            try:
                subprocess.Popen(restart_command, shell=True)
                print(f"Restarting the bot with command: {restart_command}")
            except Exception:
                # cannot use logger here, the logger is stopped
                print(f"Failed to restart the bot with command: {restart_command}", file=sys.stderr)
                raise

        if self._normal_exit_flag:
            print("Stopped gracefully.")
            # shutdown the stdout and stderr since the aiormq will still be printing rabbish
            sys.stdout.close()
            sys.stderr.close()
        else:
            print("Stopped unexpectedly.", file=sys.stderr)

    def true_stop(self, *args, **kwargs):
        self._normal_exit_flag = True
        self.application.stop_running()

    def signal_stop(self, sig, *args, **kwargs):
        global _PROGRAM_SHUTDOWN_STARTED
        if _PROGRAM_SHUTDOWN_STARTED:
            return
        _PROGRAM_SHUTDOWN_STARTED = True
        if sig == signal.SIGTERM or sig == signal.SIGABRT:
            self._exit_fast = True
        self.true_stop()

    @command_callback_wrapper
    async def stop(self, u, c):
        """
        stop - stop the bot
        Stop the bot.
        """
        self.check(CheckLevel.MASTER)
        self.true_stop()

    def get_module(self, top_name_or_clsss: Union[str, Type[_T]]) -> Optional[_T]:
        if isinstance(top_name_or_clsss, str):
            return self._module_keeper.get_module(top_name_or_clsss)  # type: ignore
        return self._module_keeper.get_module_by_class(top_name_or_clsss)

    def remove_job_if_exists(self, name: str) -> bool:
        """Remove job with given name. Returns whether job was removed."""
        current_jobs = self.job_queue.get_jobs_by_name(name)
        if not current_jobs:
            return False
        for job in current_jobs:
            job.schedule_removal()
        return True

    @command_callback_wrapper
    async def debug_mode(self, update, context):
        """
        debug_mode - switch debug mode on or off
        Switch debug mode on/off.
        """
        self.check(CheckLevel.MASTER)
        root_logger = get_root_logger()
        old_level = root_logger.level
        if old_level == LOGLEVEL_DEBUG:
            if self._old_log_level is None:
                raise RuntimeError("Invalid state: maybe default debug level is DEBUG!")
            self.debug_info("debug off")
            root_logger.setLevel(self._old_log_level)
            self._old_log_level = None
            await self.reply(Lang.t(Lang.DEBUG_MODE_OFF))
        else:
            self._old_log_level = old_level
            root_logger.setLevel(LOGLEVEL_DEBUG)
            self.debug_info("debug on")
            await self.reply(Lang.t(Lang.DEBUG_MODE_ON))

    async def _internal_exec(self, command: str):
        if not command:
            await self.error_info(Lang.t(Lang.NO_EXEC_COMMAND))
            return
        codes = command.split("\n")
        try:
            code_string = _INTERNAL_TEST_EXEC_COMMAND_PREFIX + ''.join(f'\n    {l}' for l in codes)
            _LOGGER.warning("executing: %s", code_string)
            exec(code_string)  # pylint: disable=exec-used
            ans = await locals()["__t"]()
        except Exception:
            asyncio.get_running_loop().create_task(self.reply(Lang.t(Lang.EXEC_FAILED)))
            raise
        await self.reply(Lang.t(Lang.EXEC_SUCCEEDED).format(ans), parse_mode="MarkdownV2")

    @command_callback_wrapper
    async def exec(self, update: Update, context: RichCallbackContext):
        """
        exec - execute python code
        Execute python code.
        `antares_bot.test_commands` is wildcard imported by default.
        You can use `self` to refer to the bot instance.
        """
        self.check(CheckLevel.MASTER)
        if context.args is None or len(context.args) == 0:
            # if is a reply
            try:
                reply_to = update.message.reply_to_message.text.strip()  # type: ignore
            except AttributeError:
                await self.error_info(Lang.t(Lang.NO_EXEC_COMMAND))
                return
            await self._internal_exec(reply_to)
            return
        assert update.message is not None and update.message.text is not None
        command = self.get_message_after_command(update)
        await self._internal_exec(command)

    @classmethod
    def data_dir(cls):
        return os.path.join(os.path.curdir, read_user_cfg(BasicConfig, "DATA_DIR"))

    async def _daily_job(self, context: RichCallbackContext):
        is_debug = self._is_debug_level()
        if is_debug:
            _debug_ids = []
        for old_id in self.callback_manager.history.pop_before_keys(time.time() - TIME_IN_A_DAY):
            self.callback_manager.pop_data(old_id)
            if is_debug:
                _debug_ids.append(old_id)
        if is_debug:
            _LOGGER.debug("Daily job: removed %s keys from callback manager", _debug_ids)
        #
        _LOGGER.warning("Start running daily jobs for each module")
        with ContextReverseHelper():
            await asyncio.gather(*(module.module_instance.daily_job() for module in self._module_keeper.get_all_enabled_modules() if module.module_instance is not None))

    @command_callback_wrapper
    async def get_id(self, update: Update, context: RichCallbackContext):
        """
        get_id - get the user id / chat id
        Get the user id / chat id.
        If replying to a forwarded message, get the id information of the forward prigin.
        Otherwise:
        - In a group chat:
          - If replying to a message, get the user id of the replied message;
          - otherwise, get the id of you and the group.
        - In a private chat: get your id.
        """
        if update.message is not None and update.message.reply_to_message is not None:
            message = update.message.reply_to_message
            forward_origin = message.forward_origin
            if forward_origin is not None:
                if forward_origin.type == forward_origin.CHANNEL:
                    forward_origin_channel = cast(MessageOriginChannel, forward_origin)
                    return await self.reply(
                        Lang.t(Lang.FORWARD_MESSAGE_FROM_CHANNEL).format(forward_origin_channel.chat.id),
                        parse_mode="MarkdownV2"
                    )
                elif forward_origin.type == forward_origin.CHAT:
                    forward_origin_chat = cast(MessageOriginChat, forward_origin)
                    return await self.reply(
                        Lang.t(Lang.FORWARD_MESSAGE_FROM_GROUP).format(forward_origin_chat.sender_chat.id),
                        parse_mode="MarkdownV2"
                    )
                elif forward_origin.type == forward_origin.USER:
                    forward_origin_user = cast(MessageOriginUser, forward_origin)
                    return await self.reply(
                        Lang.t(Lang.FORWARD_MESSAGE_FROM_USER).format(forward_origin_user.sender_user.id),
                        parse_mode="MarkdownV2"
                    )
                elif forward_origin.type == forward_origin.HIDDEN_USER:
                    return await self.reply(Lang.t(Lang.FORWARD_FROM_HIDDEN_USER))
        if context.is_group_chat() and update.message is not None and update.message.reply_to_message is not None and update.message.reply_to_message.from_user is not None:
            return await self.reply(
                f"{Lang.t(Lang.GROUP_ID).format(context.chat_id)}\n{Lang.t(Lang.REPLY_MESSAGE_USER_ID).format(update.message.reply_to_message.from_user.id)}",
                parse_mode="MarkdownV2"
            )
        elif context.is_group_chat():
            return await self.reply(
                f"{Lang.t(Lang.GROUP_ID).format(context.chat_id)}\n{Lang.t(Lang.USER_ID).format(context.user_id)}",
                parse_mode="MarkdownV2"
            )
        else:
            return await self.reply(
                Lang.t(Lang.USER_ID).format(context.user_id),
                parse_mode="MarkdownV2"
            )

    async def _internal_full_help(self, context: RichCallbackContext):
        ret = ""
        for command in self.handler_docs.keys():
            ret += f"`/help {command}`\n"
        await self.success_info(ret, parse_mode="Markdown")

    @classmethod
    def _match_helpdoc_line0_command_list_format(cls, command: str, doc: str, is_extract=True):
        doc = doc.strip()
        _left, _, _right = doc.partition("\n")
        line0 = _left
        command_prefix = command + " - "
        if line0.startswith(command_prefix):
            return line0 if is_extract else _right
        return None if is_extract else doc

    def _internal_generate_command_list(self) -> str:
        content = []
        for command, doc in self.handler_docs.items():
            tmp = self._match_helpdoc_line0_command_list_format(command, doc)
            if tmp is not None:
                content.append(tmp)
        if len(content) == 0:
            return ""
        content.sort()
        content = ["```"] + content + ["```"]
        return '\n'.join(content)

    @command_callback_wrapper
    async def help(self, update: Update, context: RichCallbackContext):
        """
        help - show command helps
        `/help [command]`: get the docstring for commands.
        `/help to-command-list`: get the command list for setting up at BotFather.
        """
        assert context.args is not None
        if len(context.args) == 0:
            return await self._internal_full_help(context)

        command = context.args[0]
        if command == "to-command-list":
            content = self._internal_generate_command_list()
            if content:
                return await self.reply(content, parse_mode="Markdown")
            else:
                return await self.error_info("No command list available")
        doc = self.handler_docs.get(command)
        if doc is None:
            return await self.error_info(Lang.t(Lang.NO_SUCH_COMMAND).format(command))
        doc = self._match_helpdoc_line0_command_list_format(command, doc, is_extract=False)
        if doc is not None:
            doc = trim_spaces_before_line(doc)
        if not doc:
            doc = "No doc"
        return await self.success_info(f"/{markdown_escape(command)}:\n{doc}", parse_mode="Markdown")

    @command_callback_wrapper
    async def restart(self, update: Update, context: RichCallbackContext):
        """
        Restart the bot.
        """
        self.check(CheckLevel.MASTER)
        self._post_stop_restart_flag = True
        self.true_stop()


__bot_singleton = None


def destroy_bot_instance():
    global __bot_singleton
    if __bot_singleton is not None:
        __bot_singleton.stop()
        __bot_singleton = None


__REGISTERED_BOT_CLASS = TelegramBot


def register_bot_class(cls: Type[TelegramBot]):
    if issubclass(cls, TelegramBot):
        global __REGISTERED_BOT_CLASS
        __REGISTERED_BOT_CLASS = cls
    else:
        raise RuntimeError("Invalid bot class")


def get_bot_instance() -> TelegramBot:
    global __bot_singleton
    if __bot_singleton is None:
        __bot_singleton = __REGISTERED_BOT_CLASS()
    return __bot_singleton



# pylint: disable=no-member, not-callable, arguments-differ
import re
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Pattern, Type, Union, overload

from telegram import Update
from telegram.ext import CallbackQueryHandler, CommandHandler, MessageHandler
from telegram.ext import filters as filters_module

from antares_bot.basic_language import BasicLanguage as L
from antares_bot.bot_logging import get_logger
from antares_bot.context_manager import ContextHelper
from antares_bot.error import InvalidChatTypeException, InvalidQueryException, UserPermissionException, permission_exceptions


if TYPE_CHECKING:
    from telegram.ext import BaseHandler

    from antares_bot.context import RichCallbackContext

_LOGGER = get_logger(__name__)


class CallbackBase(object):
    """
    Base class for all callback wrappers.
    subclasses need to implement `kwargs` property and `on_init`.
    """
    handler_type: Type["BaseHandler"]

    def __init__(self, func, *args, **kwargs):
        self._instance = None
        self.on_init(*args, **kwargs)
        self._register_and_wrap(func)
        self._pre_executer = None

    def on_init(self, *args, **kwargs):
        raise NotImplementedError

    def _register_and_wrap(self, func):
        wraps(func)(self)

    async def __call__(self, update: Update, context: "RichCallbackContext"):
        # pre execute
        await self.pre_execute(update, context)

        # check blacklist
        # pass
        with ContextHelper(context):
            try:
                if self._instance is not None:
                    return await self.__wrapped__(self._instance, update, context)  # type: ignore
                else:
                    return await self.__wrapped__(update, context)  # type: ignore
            except permission_exceptions() as e:
                try:
                    if self.handler_type == CommandHandler and not context.is_channel_message():
                        if isinstance(e, UserPermissionException):
                            from antares_bot.bot_inst import get_bot_instance
                            await get_bot_instance().reply(L.t(L.NO_PERMISSION))
                        elif isinstance(e, InvalidChatTypeException):
                            from antares_bot.bot_inst import get_bot_instance
                            await get_bot_instance().reply(L.t(L.INVALID_CHAT_TYPE).format(context.chat_type_str()))
                except Exception:
                    _LOGGER.error("%s.__call__", self.__class__.__name__, exc_info=True)
            except InvalidQueryException as e:
                _LOGGER.warning("Invalid query: %s %s", update.callback_query, e)

    def __get__(self, instance, cls):
        if instance is not None:
            self._instance = instance
        return self

    def to_handler(self, **kwds):
        kwds.update(self.kwargs)
        return self.handler_type(callback=self, **kwds)

    async def pre_execute(self, update: Update, context: "RichCallbackContext"):
        if self._pre_executer:
            await self._pre_executer(update, context)

    def __repr__(self) -> str:
        try:
            name = self.__name__  # type: ignore
        except AttributeError:
            name = self.__class__.__name__
        return f"{name}, of type {self.handler_type.__name__}"


class CommandCallback(CallbackBase):
    handler_type = CommandHandler

    def on_init(self, filters, block):
        self.filters = filters
        self.block = block

    @property
    def kwargs(self):
        return {
            "filters": self.filters,
            "block": self.block,
            "command": self.__wrapped__.__name__,  # pylint: disable=no-member
        }


class GeneralCallback(CallbackBase):
    PRE_EXUCUTER_KW = 'pre_executer'

    def on_init(self, handler_type, kwargs: dict):
        self.handler_type = handler_type
        self.kwargs = kwargs
        #
        pre_executer = kwargs.pop(self.PRE_EXUCUTER_KW, None)
        if pre_executer is not None:
            self._pre_executer = pre_executer


class _CommandCallbackMethodDecor(object):
    """
    Internal decorator for command callback functions.
    """

    def __init__(
        self,
        filters: Optional[filters_module.BaseFilter] = None,
        block: bool = False
    ):
        self.filters = filters
        self.block = block

    def __call__(self, func):
        return CommandCallback(func, self.filters, self.block)


class GeneralCallbackWrapper(object):
    """
    Internal decorator for command callback functions.
    """

    def __init__(
        self,
        handler_type, **kwargs
    ):
        self.handler_type = handler_type
        self.kwargs = kwargs

    def __call__(self, func):
        return GeneralCallback(func, self.handler_type, self.kwargs)


class ConditionFilter(filters_module.BaseFilter):
    def __init__(self, condition: Callable[[Update], bool]):
        super().__init__()
        self.condition = condition

    def check_update(self, update: Update) -> bool:
        return self.condition(update)


@overload
def command_callback_wrapper(func: Callable) -> CommandCallback:
    ...


@overload
def command_callback_wrapper(
    block: bool = False,
    filters: Optional[filters_module.BaseFilter] = None,
) -> CommandCallback:
    ...


def command_callback_wrapper(  # type: ignore
    block: Any = False,
    filters: Optional[filters_module.BaseFilter] = None,
):
    if callable(block):
        return _CommandCallbackMethodDecor()(block)
    return _CommandCallbackMethodDecor(filters, block)


def general_callback_wrapper(handler_type, block=False, **kwargs):
    if not callable(handler_type):
        raise TypeError("general_callback_wrapper use first argument to identify handler type")
    if 'block' not in kwargs:
        kwargs['block'] = block
    return GeneralCallbackWrapper(handler_type, **kwargs)


async def _btn_pre_executer(update: Update, context: "RichCallbackContext"):
    query = update.callback_query
    assert query is not None and query.data is not None
    await query.answer()


def btn_click_wrapper(
        pattern: Optional[Union[str, Pattern[str], type, Callable[[object], Optional[bool]]]] = None
):
    if isinstance(pattern, str):
        # startswith `pattern`
        pattern = re.compile(f"^{pattern}")
    kwargs: Dict[str, Any] = {}
    kwargs['pattern'] = pattern
    kwargs[GeneralCallback.PRE_EXUCUTER_KW] = _btn_pre_executer
    return general_callback_wrapper(CallbackQueryHandler, **kwargs)


@overload
def msg_handle_wrapper(filters: Callable[["Update"], Any]) -> GeneralCallbackWrapper:
    ...


@overload
def msg_handle_wrapper(func: Callable[[Any, "Update", "RichCallbackContext"], Any]) -> GeneralCallback:
    ...


@overload
def msg_handle_wrapper(filters: Optional[filters_module.BaseFilter] = None) -> GeneralCallbackWrapper:
    ...


def msg_handle_wrapper(*args, **kwargs):
    """
    message handler wrapper.
    if need check before execute, pass a filter object in kwargs `filters`
    """
    if len(args) > 0 and callable(args[0]):
        return general_callback_wrapper(MessageHandler, filters=None)(args[0])
    # has kwargs
    filters = kwargs.get('filters', None)
    if filters is not None and not isinstance(filters, filters_module.BaseFilter):
        # wrap it
        _filters = ConditionFilter(filters)
        kwargs['filters'] = _filters
    return general_callback_wrapper(MessageHandler, *args, **kwargs)


photo_handle_wrapper = general_callback_wrapper(MessageHandler, filters=filters_module.PHOTO & (~filters_module.ChatType.CHANNEL))



import importlib
import os
import sys
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, List, Optional, Type, TypeVar, cast

from antares_bot.bot_default_cfg import AntaresBotConfig
from antares_bot.bot_logging import get_logger
from antares_bot.module_base import TelegramBotModuleBase
from antares_bot.utils import read_user_cfg


if TYPE_CHECKING:
    from telegram.ext import Application

    from antares_bot.bot_inst import TelegramBot

_T = TypeVar("_T", bound=TelegramBotModuleBase)

MODULE_PRIORITY_STR = "MODULE_PRIORITY"
VALID_MODULE_RANGE = (0, 256)
DEFAULT_PRIORITY = 128

_LOGGER = get_logger(__name__)


class TelegramBotModuleDesc(Generic[_T]):
    """
    A descriptor of a tgbot module (usually a class, not a py module).
    """

    def __init__(self, top_name: str, kls: Type[_T]) -> None:
        self.top_name = top_name
        self.kls = kls
        self.module_instance: Optional[_T] = None
        self._enabled = True

    @property
    def priority(self) -> int:
        return getattr(self.kls, MODULE_PRIORITY_STR, DEFAULT_PRIORITY)

    @property
    def enabled(self) -> bool:
        return self._enabled

    def check_priority_valid(self) -> None:
        l, r = VALID_MODULE_RANGE
        priority = self.priority
        if priority < l or priority >= r:
            raise ValueError(f"{MODULE_PRIORITY_STR} of {self.top_name} is invalid, should be in [{l}, {r})")

    def do_init(self, parent: "TelegramBot") -> None:
        self.module_instance = self.kls(parent)
        self.module_instance.do_init()

    async def post_init(self, app: "Application") -> None:
        if self.module_instance is not None:
            await self.module_instance.post_init(app)

    async def do_stop(self):
        if self.module_instance is not None:
            await self.module_instance.do_stop()

    def set_enabled(self, enabled: bool) -> None:
        self._enabled = enabled

    def py_module(self):
        return sys.modules[self.kls.__module__]

    def __repr__(self) -> str:
        return f"TelegramBotModule: {self.top_name}"


class ModuleKeeper(object):
    _STR = 1
    _TYPE = 2

    def __init__(self) -> None:
        self._ordered_modules: List[TelegramBotModuleDesc[TelegramBotModuleBase]] = []
        self._modules_dict: Dict[str, TelegramBotModuleDesc[TelegramBotModuleBase]] = dict()
        self._class2module_dict: Dict[Type[TelegramBotModuleBase], TelegramBotModuleDesc[TelegramBotModuleBase]] = dict()
        self._disabled_modules_dict: Dict[str, TelegramBotModuleDesc[TelegramBotModuleBase]] = dict()
        self._disabled_class2module_dict: Dict[Type[TelegramBotModuleBase], TelegramBotModuleDesc[TelegramBotModuleBase]] = dict()

    def load_all(self) -> None:
        """
        Should only be called once at init.
        """
        self._sort_and_set_modules(self._import_all_modules())

    def reload_all(self) -> None:
        self.clear()
        self.load_all()

    def get_module(self, top_name: str) -> Optional[TelegramBotModuleBase]:
        py_module = self._find_module_internal(top_name, self._STR)
        if py_module is not None:
            return py_module.module_instance
        return None

    def get_module_by_class(self, cls: Type[_T]) -> Optional[_T]:
        py_module = self._find_module_internal(cls, self._TYPE)
        if py_module is not None:
            return cast(_T, py_module.module_instance)
        return None

    def get_all_enabled_modules(self) -> List[TelegramBotModuleDesc[TelegramBotModuleBase]]:
        return self._ordered_modules

    def run_over(self, func: Callable[[TelegramBotModuleDesc[TelegramBotModuleBase]], Any]):
        for module in self._ordered_modules:
            func(module)

    def disable_module(self, top_name: str) -> None:
        self._disable_module_internal(top_name, self._STR)

    def disable_module_by_class(self, cls: Type[_T]) -> None:
        self._disable_module_internal(cls, self._TYPE)

    def clear(self) -> None:
        del self._ordered_modules[:]
        self._modules_dict.clear()
        self._class2module_dict.clear()

    def _find_module_internal(self, k, _type: int) -> Optional[TelegramBotModuleDesc[TelegramBotModuleBase]]:
        if _type == self._STR:
            return self._find_module_from(self._modules_dict, self._disabled_modules_dict, k)
        return self._find_module_from(self._class2module_dict, self._disabled_class2module_dict, k)

    @staticmethod
    def _find_module_from(d1: dict, d2: dict, k) -> Optional[TelegramBotModuleDesc[TelegramBotModuleBase]]:
        py_module = d1.get(k, None)
        if py_module is not None:
            return py_module
        return d2.get(k, None)

    def _disable_module_internal(self, k, _type: int) -> None:
        if _type == self._STR:
            module = self._disable_module_from(self._modules_dict, self._disabled_modules_dict, k)
        else:
            module = self._disable_module_from(self._class2module_dict, self._disabled_class2module_dict, k)
        if module is not None:
            self._remove_from_sorted_modules(module)
        raise ValueError(f"Module {k} not found")

    @staticmethod
    def _disable_module_from(
        d1: Dict[Any, TelegramBotModuleDesc],
        d2: Dict[Any, TelegramBotModuleDesc],
        k
    ) -> Optional[TelegramBotModuleDesc[TelegramBotModuleBase]]:
        # TODO
        raise NotImplementedError
        # module = d1.get(k, None)
        # if module is not None:
        #     module.set_enabled(False)
        #     del d1[k]
        #     d2[k] = module
        #     return module
        # return None

    def _remove_from_sorted_modules(self, module: TelegramBotModuleDesc[TelegramBotModuleBase]) -> None:
        self._ordered_modules.remove(module)

    @staticmethod
    def _import_all_modules() -> Dict[str, Type[TelegramBotModuleBase]]:

        def _module_check(_module, _module_top_name: str, module_store_name: str):
            _names = _module_top_name.split("_")
            class_name = ''.join([name.capitalize() for name in _names])
            kls = getattr(_module, class_name, None)
            if not isinstance(kls, type):
                return None
            try:
                if not issubclass(kls, TelegramBotModuleBase):  # type: ignore
                    return None
            except Exception:
                _LOGGER.error("%s is not a subclass of TelegramBotModuleBase", module_store_name)
                return None
            return kls

        def _import_module(_module_full_name: str):
            try:
                if exist_module := sys.modules.get(_module_full_name):
                    exist = True
                    module = exist_module
                else:
                    exist = False
                    module = importlib.import_module(_module_full_name)
            except Exception as e:
                if read_user_cfg(AntaresBotConfig, "IGNORE_IMPORT_MODULE_ERROR"):
                    _LOGGER.error(e)
                    return None
                else:
                    raise
            return exist, module

        skip_load_formatter = "SKIP_LOAD_MODULE_{}"
        skip_load_internal_formatter = "SKIP_LOAD_INTERNAL_MODULE_{}"
        ret: Dict[str, Type[TelegramBotModuleBase]] = dict()
        cur_path = os.path.dirname(os.path.abspath(__file__))
        cur_path_folder_name = os.path.basename(cur_path)

        def _load_up(_filename: str, is_internal: bool = False, _dirname: str | None = None):
            # is_internal: e.g. internal_modules/test.py -> test
            # not is_internal: e.g. modules/test.py -> test
            # not is_internal: e.g. modules/sub_dir/sub_test.py -> sub_test
            module_top_name = _filename[:-3]
            # use cfg to control whether to load (internal) modules
            skip = read_user_cfg(AntaresBotConfig, (skip_load_internal_formatter if is_internal else skip_load_formatter).format(module_top_name.upper())) or False
            if skip:
                return
            # is_internal: e.g. internal_modules/test.py -> internal_modules.test
            # not is_internal: same as `module_top_name`
            if is_internal:
                module_store_name = "internal_modules." + module_top_name
            else:
                module_store_name = module_top_name

            if module_store_name in ret:
                _LOGGER.error("%s is duplicated", module_store_name)
                return
            if is_internal:
                # e.g. internal_modules.test -> {cur_path_folder_name}.internal_modules.test
                module_full_name = f"{cur_path_folder_name}.{module_store_name}"
            else:
                # e.g. test.py -> modules.test
                assert _dirname is not None
                module_full_name = os.path.join(_dirname, _filename).replace(os.path.sep, ".")[:-3]
            # load it
            _import_result = _import_module(module_full_name)
            if _import_result is None:
                return
            exists, module = _import_result
            # check
            kls = _module_check(module, module_top_name, module_store_name)
            if kls is None:
                return
            # finalize
            if not exists:
                _LOGGER.info("loaded module %s", module_store_name)
            # is_internal: e.g. internal_modules/test.py, internal_modules.test -> Test
            # not is_internal: e.g. modules/test.py, test -> Test
            # not is_internal: e.g. modules/sub_dir/sub_test.py, sub_test -> SubTest
            ret[module_store_name] = kls

        # first load the internal modules

        os.path.join(cur_path, "internal_modules")
        if read_user_cfg(AntaresBotConfig, "SKIP_LOAD_ALL_INTERNAL_MODULES"):
            _LOGGER.warning("SKIP_LOAD_ALL_INTERNAL_MODULES is set to True, no internal modules will be loaded")
        else:
            for filename in os.listdir(os.path.join(cur_path, "internal_modules")):
                if filename.endswith(".py") and filename != "__init__.py":
                    _load_up(filename, is_internal=True)
                    continue

        # load user modules
        for dirname, _, filenames in os.walk("modules"):
            if dirname.endswith("__pycache__"):
                continue
            for filename in filenames:
                if filename.endswith(".py") and filename != "__init__.py":
                    _load_up(filename, _dirname=dirname)
                    continue

        return ret

    @staticmethod
    def _sort_modules(klss: Dict[str, Type[TelegramBotModuleBase]]):
        modules: List[TelegramBotModuleDesc[TelegramBotModuleBase]] = []
        temp_dict: defaultdict[int, List[TelegramBotModuleDesc[TelegramBotModuleBase]]] = defaultdict(list)
        for top_name, kls in klss.items():
            module = TelegramBotModuleDesc(top_name, kls)
            module.check_priority_valid()
            temp_dict[module.priority].append(module)
        for lst in temp_dict.values():
            lst.sort(key=lambda x: x.top_name)

        for k in sorted(temp_dict.keys()):
            modules.extend(temp_dict[k])
        return modules

    def _sort_and_set_modules(self, klss: Dict[str, Type[TelegramBotModuleBase]]):
        sorted_modules = self._sort_modules(klss)
        #
        self._add_modules(sorted_modules)

    def _add_module(self, module: TelegramBotModuleDesc[TelegramBotModuleBase]):
        self._maintain_add_module_internal(module)

    def _add_modules(self, modules: List[TelegramBotModuleDesc[TelegramBotModuleBase]]):
        for module in modules:
            self._maintain_add_module_internal(module)

    def _maintain_add_module_internal(self, module: TelegramBotModuleDesc[TelegramBotModuleBase]):
        if module.enabled:
            self._ordered_modules.append(module)
            self._modules_dict[module.top_name] = module
            self._class2module_dict[module.kls] = module
        else:
            self._disabled_modules_dict[module.top_name] = module
            self._disabled_class2module_dict[module.kls] = module




class BaseBotFrameworkException(Exception):
    pass


class InvalidQueryException(BaseBotFrameworkException):
    pass


class UserPermissionException(BaseBotFrameworkException):
    pass


class IgnoreChannelUpdateException(BaseBotFrameworkException):
    pass


class InvalidChatTypeException(BaseBotFrameworkException):
    pass


def permission_exceptions():
    return (
        UserPermissionException,
        IgnoreChannelUpdateException,
        InvalidChatTypeException,
    )



from enum import Enum
from typing import TYPE_CHECKING

from antares_bot.bot_default_cfg import BasicConfig
from antares_bot.utils import read_user_cfg


if TYPE_CHECKING:
    from antares_bot.context import RichCallbackContext


class CheckLevel(Enum):
    MASTER = 0
    ADMIN = 1
    USER = 2
    ANY = 3


class ConditionLimit(Enum):
    GROUP = 1
    PRIVATE = 2
    CHANNEL = 4
    CALLBACK_QUERY = 8
    CHAT = GROUP | PRIVATE
    ALL = 0xf


class PermissionState(Enum):
    PASSED = 0
    INVALID_USER = 1
    INVALID_CHAT_TYPE = 2
    IGNORE_CHANNEL = 3


def permission_check(context: "RichCallbackContext", level: CheckLevel, limit: ConditionLimit = ConditionLimit.ALL):
    if context.is_private_chat():
        if 0 == (limit.value & ConditionLimit.PRIVATE.value):
            return PermissionState.INVALID_CHAT_TYPE
    elif context.is_group_chat():
        if 0 == (limit.value & ConditionLimit.GROUP.value):
            return PermissionState.INVALID_CHAT_TYPE
    elif context.is_channel_message():
        if 0 == (limit.value & ConditionLimit.CHANNEL.value):
            return PermissionState.IGNORE_CHANNEL
    if level == CheckLevel.MASTER:
        master_id = read_user_cfg(BasicConfig, "MASTER_ID")
        is_master = context.chat_id == master_id or context.user_id == master_id
        return PermissionState.PASSED if is_master else PermissionState.INVALID_USER
    # add more checks in future
    return PermissionState.PASSED



import bisect
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, Iterable, List, Optional, Tuple, TypeVar, Union

from telegram import InlineKeyboardButton

from antares_bot.utils import flatten_button


if TYPE_CHECKING:
    from telegram import InlineKeyboardMarkup

_DataType = TypeVar("_DataType")


class CallbackHistoryManager:
    __slots__ = ("history_queue")

    def __init__(self) -> None:
        self.history_queue: List[Tuple[float, int]] = []

    def enqueue(self, key: int) -> None:
        self.history_queue.append((time.time(), key))

    def pop_before_keys(self, before: float):
        history_queue = self.history_queue
        idx = bisect.bisect_left(history_queue, (before, 0))
        if idx > 0:
            self.history_queue = history_queue[idx:]
        yield from (k for _, k in history_queue[:idx])


class CallbackDataManager:
    __slots__ = ("id", "_dict", "history")

    def __init__(self) -> None:
        self.id = 0
        self._dict: Dict[int, Any] = {}
        self.history = CallbackHistoryManager()

    def set_data(self, data=None) -> str:
        """
        store data and return the key to retrieve it later.
        the key is a string of int.
        """
        if data is not None:
            self._dict[self.id] = data
            self.history.enqueue(self.id)
        self.id += 1
        return str(self.id - 1)

    def pop_data(self, _id: Union[str, int]) -> Any:
        """
        pop the data by the key.
        """
        n_id = int(_id)
        return self._dict.pop(n_id, None)

    def peek_data(self, _id: Union[str, int]) -> Any:
        """
        peek the data by the key.
        """
        n_id = int(_id)
        return self._dict.get(n_id, None)

    def modify_data(self, _id: Union[str, int], data: Any) -> None:
        """
        modify the data by the key.
        """
        n_id = int(_id)
        if data is not None:
            self._dict[n_id] = data
        else:
            self._dict.pop(n_id, None)


class PersistKeyboards(Generic[_DataType]):
    """
    Parameter of repr_cb: idx, cb_data_key, data

    The normal usage is like:
    * construct with callback data manager.
    * call `setup_use_data()` to store all data into self, and also the repr_cb, which is used to generate the text of buttons. If you want to customize, you can split this method into two steps: `store_data()`, `setup_use_keys()`.
    * call `get_reply_markup()` to get the InlineKeyboardMarkup.

    When the data is retrieved:
    * the type of data is `Tuple[_DataType, PersistKeyboards[_DataType]]`.
    * the index can be retrieved by calling `self.idx(get_cb_data_key)`, where `get_cb_data_key = bot_module._get_cb_data_key(query)`.
    """

    def __init__(self, cb_manager: CallbackDataManager) -> None:
        self.cb_data_keys: List[str] = []
        self.repr_cb: Optional[Callable[[int, str, _DataType], str]] = None
        self.cb_manager = cb_manager
        self._idx_map: Dict[str, int] = dict()

    def get_reply_markup(self, pattern_key: str, button_in_row: int) -> Optional["InlineKeyboardMarkup"]:
        """
        get the InlineKeyboardMarkup. return `None` if no callback data keys are set.
        """
        if len(self.cb_data_keys) == 0:
            return None
        return flatten_button(self._to_buttons(pattern_key), button_in_row)

    def _to_buttons(self, pattern_key: str) -> List[InlineKeyboardButton]:
        return [InlineKeyboardButton(self._get_text(i), callback_data=f"{pattern_key}:{k}") for i, k in enumerate(self.cb_data_keys)]

    def setup_use_data(self, data_list: Iterable[_DataType], repr_cb: Optional[Callable[[int, str, _DataType], str]] = None):
        """
        setting up the keyboard.
        store all callback data into callback manager, store corresponding keys into persist keyboard, and set the `repr_cb` for formatting the button text.

        Parameter of repr_cb: `idx`, `cb_data_key`, `data`.
        """
        cb_data_keys: List[str] = []
        for data in data_list:
            cb_data_keys.append(self.store_data(data))
        self.setup_use_keys(cb_data_keys, repr_cb)

    def setup_use_keys(self, cb_data_keys: List[str], repr_cb: Optional[Callable[[int, str, _DataType], str]] = None):
        """
        setting up the keyboard.
        store all callback data keys into persist keyboard, and set the `repr_cb` for formatting the button text.

        Parameter of repr_cb: `idx`, `cb_data_key`, `data`.
        """
        self.cb_data_keys = cb_data_keys
        self.repr_cb = repr_cb
        for i, k in enumerate(cb_data_keys):
            self._idx_map[k] = i

    def modify_data(self, cb_data_key: str, data: _DataType) -> None:
        idx = self.idx(cb_data_key)
        self.modify_data_by_index(idx, data)

    def modify_data_by_index(self, idx: int, data: _DataType) -> None:
        cb_data_key = self.cb_data_keys[idx]
        self.cb_manager.modify_data(cb_data_key, (data, self))

    def get_data_by_index(self, idx: int) -> _DataType:
        cb_data_key = self.cb_data_keys[idx]
        data: _DataType = self.cb_manager.peek_data(cb_data_key)[0]
        return data

    def _get_text(self, idx: int) -> str:
        if self.repr_cb is None:
            return str(idx)
        cb_data_key = self.cb_data_keys[idx]
        data: _DataType = self.cb_manager.peek_data(cb_data_key)[0]
        return self.repr_cb(idx, cb_data_key, data)

    def idx(self, cb_data_key: str) -> int:
        return self._idx_map[cb_data_key]

    def store_data(self, data: _DataType) -> str:
        return self.cb_manager.set_data((data, self))

    def clean(self) -> None:
        """
        clean all refs in callback manager when the keyboard is no longer used
        """
        for k in self.cb_data_keys:
            self.cb_manager.pop_data(k)
        self.cb_data_keys = []
        self._idx_map.clear()
        self.repr_cb = None

    def __len__(self) -> int:
        return len(self.cb_data_keys)



import asyncio
import logging
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Tuple, Union, cast

from telegram import Message
# from antares_bot.patching.conversation_handler_ex import ConversationHandlerEx
from telegram.ext import ConversationHandlerEx

from antares_bot.basic_language import BasicLanguage as L
from antares_bot.bot_base import TelegramBotBase
from antares_bot.error import InvalidQueryException
from antares_bot.framework import command_callback_wrapper
from antares_bot.utils import exception_manual_handle


if TYPE_CHECKING:
    from typing import Self

    from telegram import CallbackQuery, InlineKeyboardMarkup
    from telegram.ext import Application, BaseHandler

    from antares_bot.bot_inst import TelegramBot
    from antares_bot.callback_manager import PersistKeyboards
    from antares_bot.context import RichCallbackContext
    from antares_bot.framework import CallbackBase


class TelegramBotModuleBase(TelegramBotBase):
    if TYPE_CHECKING:
        INST: Any

    @property
    def bot_id(self):
        return self.parent.bot_id

    @classmethod
    def get_inst(cls) -> "Self":
        return cast("Self", cls.INST)

    def __init__(self, parent: "TelegramBot") -> None:
        self.parent = parent
        self._register_inst()
        self._handlers: Optional[List[Union["CallbackBase", "BaseHandler"]]] = None

    def do_init(self) -> None:
        ...

    async def post_init(self, app: "Application") -> None:
        ...

    async def do_stop(self) -> None:
        ...

    async def daily_job(self) -> None:
        """
        Run at 0 o'clock every day.
        """

    def _register_inst(self):
        self.__class__.INST = self

    def collect_handlers(self):
        if self._handlers is None:
            self._handlers = self.mark_handlers()
        return self._handlers

    def mark_handlers(self) -> List[Union["CallbackBase", "BaseHandler"]]:
        """Override to mark all handlers that will be collected."""
        return []

    # -------------------------helper functions-------------------------
    # -------------------------DO NOT OVERRIDE--------------------------
    def make_btn_callback(self, key: str, data: Iterable) -> Tuple[List[str], List[str]]:
        """
        return a list of callback data strings, which can be used to retrieve data later.
        Need to call `cache_cb_keys_by_{}` to cache the keys after the message is sent.

        Return: cb_keys, cb_datas.
        """
        raw_keys = []
        keys = []
        for dt in data:
            key_raw = self.parent.callback_manager.set_data(dt)
            raw_keys.append(key_raw)
            keys.append(f"{key}:{key_raw}")
        return raw_keys, keys

    def _get_cb_data_key(self, query: "CallbackQuery"):
        """
        generally the data format stored in callback manager is `key:xxx`.
        get the callback data `xxx` after the colon.
        """
        assert query.data is not None
        return query.data.split(':')[1]

    def get_btn_callback_data(self, query: "CallbackQuery", pop: bool = False, check_valid=False):
        """
        get button callback data from callback manager from query.
        if `pop` is True, the data will be removed from callback manager.
        if `check_valid` is True, it will call `on_invalid_query` if the data is `None`.
        """
        k = self._get_cb_data_key(query)
        ret = self.parent.callback_manager.pop_data(k) if pop else self.parent.callback_manager.peek_data(k)
        if ret is None and check_valid:
            loop = asyncio.get_running_loop()
            loop.create_task(self.on_invalid_query(query))
            raise InvalidQueryException
        return ret

    @property
    def cancel(self):
        """
        a cancel callback for conversation handler.
        """
        async def cancel(x, y):
            await self.reply(L.t(L.CANCELLED))
            return ConversationHandlerEx.END
        return command_callback_wrapper(cancel)

    @property
    def job_queue(self):
        return self.parent.job_queue

    async def _manual_handle_exception(self, logger: logging.Logger, e: Exception):
        return await exception_manual_handle(logger, e)

    def cache_cb_keys_by_id(self, chat_id: int, msg_id: int, cb_keys: List[str]) -> None:
        """
        cache the callback keys by message id.
        note the `key` here is the key of the callback manager.
        """
        self.parent.callback_key_dict[(chat_id, msg_id)] = cb_keys

    def cache_cb_keys_by_message(self, message: Message, cb_keys: List[str]) -> None:
        """
        cache the callback keys by message id.
        note the `key` here is the key of the callback manager.
        """
        self.cache_cb_keys_by_id(message.chat_id, message.message_id, cb_keys)

    def clean_cb_keys_by_id(self, chat_id: int, msg_id: int) -> None:
        """
        clean the callback keys by message id.
        note the `key` here is the key of the callback manager.
        """
        cb_keys = self.parent.callback_key_dict.pop((chat_id, msg_id), [])
        for key in cb_keys:
            self.parent.callback_manager.pop_data(key)

    def clean_cb_keys(self, context: "RichCallbackContext") -> None:
        self.clean_cb_keys_by_id(context.chat_id, context.message_id)

    def clean_cb_keys_by_message(self, message: Message) -> None:
        self.clean_cb_keys_by_id(message.chat_id, message.message_id)

    async def query_remove_btn(self, query: "CallbackQuery", text: str = "", remove_message: bool = False, reply_markup: Optional["InlineKeyboardMarkup"] = None):
        assert query.message
        assert isinstance(query.message, Message)
        self.clean_cb_keys_by_message(query.message)
        if remove_message:
            return await query.message.delete()
        #
        return await query.message.edit_text(text, reply_markup=reply_markup)

    async def on_invalid_query(self, query):
        """
        Call when the data in callback manager is removed.
        """
        await self.query_remove_btn(query, text="这个按钮无效了呢……", remove_message=False)

    def query_at_btn_index(self, query: "CallbackQuery") -> int:
        """
        Call when the query uses a keyboard.
        """
        _, keyboard = self.get_btn_callback_data(query)
        keyboard = cast("PersistKeyboards", keyboard)
        return keyboard.idx(self._get_cb_data_key(query))
    # -----------------------helper functions end-----------------------
